# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

# SIGTERM handling based on the implementation by Virgil Dupras for dupeGuru:
# https://github.com/arsenetar/dupeguru/blob/master/run.py#L71

"""A Qt driver for TagStudio."""

import ctypes
import logging
import math
import os
import sys
import time
import typing
import webbrowser
from datetime import datetime as dt
from pathlib import Path
from queue import Queue
from typing import Optional
from PIL import Image
from PySide6 import QtCore
from PySide6.QtCore import (
    QObject,
    QThread,
    Signal,
    Qt,
    QThreadPool,
    QTimer,
    QSettings,
)
from PySide6.QtGui import (
    QGuiApplication,
    QPixmap,
    QMouseEvent,
    QColor,
    QAction,
    QFontDatabase,
    QIcon,
)
from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import (
    QApplication,
    QWidget,
    QHBoxLayout,
    QPushButton,
    QLineEdit,
    QScrollArea,
    QFileDialog,
    QSplashScreen,
    QMenu,
    QMenuBar,
    QComboBox,
)
from humanfriendly import format_timespan

from src.core.enums import SettingItems, SearchMode
from src.core.library import ItemType
from src.core.ts_core import TagStudioCore
from src.core.constants import (
    COLLAGE_FOLDER_NAME,
    BACKUP_FOLDER_NAME,
    TS_FOLDER_NAME,
    VERSION_BRANCH,
    VERSION,
    TAG_FAVORITE,
    TAG_ARCHIVED,
)
from src.core.utils.web import strip_web_protocol
from src.qt.flowlayout import FlowLayout
from src.qt.main_window import Ui_MainWindow
from src.qt.helpers.function_iterator import FunctionIterator
from src.qt.helpers.custom_runnable import CustomRunnable
from src.qt.resource_manager import ResourceManager
from src.qt.widgets.collage_icon import CollageIconRenderer
from src.qt.widgets.panel import PanelModal
from src.qt.widgets.thumb_renderer import ThumbRenderer
from src.qt.widgets.progress import ProgressWidget
from src.qt.widgets.preview_panel import PreviewPanel
from src.qt.widgets.item_thumb import ItemThumb
from src.qt.modals.build_tag import BuildTagPanel
from src.qt.modals.tag_database import TagDatabasePanel
from src.qt.modals.file_extension import FileExtensionModal
from src.qt.modals.fix_unlinked import FixUnlinkedEntriesModal
from src.qt.modals.fix_dupes import FixDupeFilesModal
from src.qt.modals.folders_to_tags import FoldersToTagsModal

# this import has side-effect of import PySide resources
import src.qt.resources_rc  # pylint: disable=unused-import

# SIGQUIT is not defined on Windows
if sys.platform == "win32":
    from signal import signal, SIGINT, SIGTERM

    SIGQUIT = SIGTERM
else:
    from signal import signal, SIGINT, SIGTERM, SIGQUIT

ERROR = f"[ERROR]"
WARNING = f"[WARNING]"
INFO = f"[INFO]"

logging.basicConfig(format="%(message)s", level=logging.INFO)


class NavigationState:
    """Represents a state of the Library grid view."""

    def __init__(
        self,
        contents,
        scrollbar_pos: int,
        page_index: int,
        page_count: int,
        search_text: str | None = None,
        thumb_size=None,
        spacing=None,
    ) -> None:
        self.contents = contents
        self.scrollbar_pos = scrollbar_pos
        self.page_index = page_index
        self.page_count = page_count
        self.search_text = search_text
        self.thumb_size = thumb_size
        self.spacing = spacing


class Consumer(QThread):
    MARKER_QUIT = "MARKER_QUIT"

    def __init__(self, queue) -> None:
        self.queue = queue
        QThread.__init__(self)

    def run(self):
        while True:
            try:
                job = self.queue.get()
                if job == self.MARKER_QUIT:
                    break
                job[0](*job[1])
            except RuntimeError:
                pass

    def set_page_count(self, count: int):
        self.page_count = count

    def jump_to_page(self, index: int):
        pass

    def nav_back(self):
        pass

    def nav_forward(self):
        pass


class QtDriver(QObject):
    """A Qt GUI frontend driver for TagStudio."""

    SIGTERM = Signal()

    preview_panel: PreviewPanel

    def __init__(self, core: TagStudioCore, args):
        super().__init__()
        self.core: TagStudioCore = core
        self.lib = self.core.lib
        self.rm: ResourceManager = ResourceManager()
        self.args = args
        self.frame_dict: dict = {}
        self.nav_frames: list[NavigationState] = []
        self.cur_frame_idx: int = -1

        self.search_mode = SearchMode.AND

        # self.main_window = None
        # self.main_window = Ui_MainWindow()

        self.branch: str = (" (" + VERSION_BRANCH + ")") if VERSION_BRANCH else ""
        self.base_title: str = f"TagStudio Alpha {VERSION}{self.branch}"
        # self.title_text: str = self.base_title
        # self.buffer = {}
        self.thumb_job_queue: Queue = Queue()
        self.thumb_threads: list[Consumer] = []
        self.thumb_cutoff: float = time.time()
        # self.selected: list[tuple[int,int]] = [] # (Thumb Index, Page Index)
        self.selected: list[tuple[ItemType, int]] = []  # (Item Type, Item ID)

        self.SIGTERM.connect(self.handleSIGTERM)

        if self.args.config_file:
            path = Path(self.args.config_file)
            if not path.exists():
                logging.warning(
                    f"[QT DRIVER] Config File does not exist creating {str(path)}"
                )
            logging.info(f"[QT DRIVER] Using Config File {str(path)}")
            self.settings = QSettings(str(path), QSettings.Format.IniFormat)
        else:
            self.settings = QSettings(
                QSettings.Format.IniFormat,
                QSettings.Scope.UserScope,
                "TagStudio",
                "TagStudio",
            )
            logging.info(
                f"[QT DRIVER] Config File not specified, defaulting to {self.settings.fileName()}"
            )

        max_threads = os.cpu_count()
        if args.ci:
            # spawn only single worker in CI environment
            max_threads = 1
        for i in range(max_threads):
            # thread = threading.Thread(target=self.consumer, name=f'ThumbRenderer_{i}',args=(), daemon=True)
            # thread.start()
            thread = Consumer(self.thumb_job_queue)
            thread.setObjectName(f"ThumbRenderer_{i}")
            self.thumb_threads.append(thread)
            thread.start()

    def open_library_from_dialog(self):
        dir = QFileDialog.getExistingDirectory(
            None, "Open/Create Library", "/", QFileDialog.ShowDirsOnly
        )
        if dir not in (None, ""):
            self.open_library(Path(dir))

    def signal_handler(self, sig, frame):
        if sig in (SIGINT, SIGTERM, SIGQUIT):
            self.SIGTERM.emit()

    def setup_signals(self):
        signal(SIGINT, self.signal_handler)
        signal(SIGTERM, self.signal_handler)
        signal(SIGQUIT, self.signal_handler)

    def start(self) -> None:
        """Launches the main Qt window."""

        loader = QUiLoader()
        if os.name == "nt":
            sys.argv += ["-platform", "windows:darkmode=2"]
        app = QApplication(sys.argv)
        app.setStyle("Fusion")
        # pal: QPalette = app.palette()
        # pal.setColor(QPalette.ColorGroup.Active,
        # 			 QPalette.ColorRole.Highlight, QColor('#6E4BCE'))
        # pal.setColor(QPalette.ColorGroup.Normal,
        # 			 QPalette.ColorRole.Window, QColor('#110F1B'))
        # app.setPalette(pal)
        home_path = Path(__file__).parent / "ui/home.ui"
        icon_path = Path(__file__).parents[2] / "resources/icon.png"

        # Handle OS signals
        self.setup_signals()
        # allow to process input from console, eg. SIGTERM
        timer = QTimer()
        timer.start(500)
        timer.timeout.connect(lambda: None)

        # self.main_window = loader.load(home_path)
        self.main_window = Ui_MainWindow(self)
        self.main_window.setWindowTitle(self.base_title)
        self.main_window.mousePressEvent = self.mouse_navigation  # type: ignore
        # self.main_window.setStyleSheet(
        # 	f'QScrollBar::{{background:red;}}'
        # 	)

        # # self.main_window.windowFlags() &
        # # self.main_window.setWindowFlag(Qt.WindowType.FramelessWindowHint, True)
        # self.main_window.setWindowFlag(Qt.WindowType.NoDropShadowWindowHint, True)
        # self.main_window.setWindowFlag(Qt.WindowType.WindowTransparentForInput, False)
        # self.main_window.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

        # self.windowFX = WindowEffect()
        # self.windowFX.setAcrylicEffect(self.main_window.winId())

        splash_pixmap = QPixmap(":/images/splash.png")
        splash_pixmap.setDevicePixelRatio(self.main_window.devicePixelRatio())
        self.splash = QSplashScreen(splash_pixmap, Qt.WindowStaysOnTopHint)  # type: ignore
        # self.splash.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.splash.show()

        if os.name == "nt":
            appid = "cyanvoxel.tagstudio.9"
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)  # type: ignore

        if sys.platform != "darwin":
            icon = QIcon()
            icon.addFile(str(icon_path))
            app.setWindowIcon(icon)

        menu_bar = QMenuBar(self.main_window)
        self.main_window.setMenuBar(menu_bar)
        menu_bar.setNativeMenuBar(True)

        file_menu = QMenu("&File", menu_bar)
        edit_menu = QMenu("&Edit", menu_bar)
        tools_menu = QMenu("&Tools", menu_bar)
        macros_menu = QMenu("&Macros", menu_bar)
        window_menu = QMenu("&Window", menu_bar)
        help_menu = QMenu("&Help", menu_bar)

        # File Menu ============================================================
        # file_menu.addAction(QAction('&New Library', menu_bar))
        # file_menu.addAction(QAction('&Open Library', menu_bar))

        open_library_action = QAction("&Open/Create Library", menu_bar)
        open_library_action.triggered.connect(lambda: self.open_library_from_dialog())
        open_library_action.setShortcut(
            QtCore.QKeyCombination(
                QtCore.Qt.KeyboardModifier(QtCore.Qt.KeyboardModifier.ControlModifier),
                QtCore.Qt.Key.Key_O,
            )
        )
        open_library_action.setToolTip("Ctrl+O")
        file_menu.addAction(open_library_action)

        save_library_action = QAction("&Save Library", menu_bar)
        save_library_action.triggered.connect(
            lambda: self.callback_library_needed_check(self.save_library)
        )
        save_library_action.setShortcut(
            QtCore.QKeyCombination(
                QtCore.Qt.KeyboardModifier(QtCore.Qt.KeyboardModifier.ControlModifier),
                QtCore.Qt.Key.Key_S,
            )
        )
        save_library_action.setStatusTip("Ctrl+S")
        file_menu.addAction(save_library_action)

        save_library_backup_action = QAction("&Save Library Backup", menu_bar)
        save_library_backup_action.triggered.connect(
            lambda: self.callback_library_needed_check(self.backup_library)
        )
        save_library_backup_action.setShortcut(
            QtCore.QKeyCombination(
                QtCore.Qt.KeyboardModifier(
                    QtCore.Qt.KeyboardModifier.ControlModifier
                    | QtCore.Qt.KeyboardModifier.ShiftModifier
                ),
                QtCore.Qt.Key.Key_S,
            )
        )
        save_library_backup_action.setStatusTip("Ctrl+Shift+S")
        file_menu.addAction(save_library_backup_action)

        file_menu.addSeparator()

        # refresh_lib_action = QAction('&Refresh Directories', self.main_window)
        # refresh_lib_action.triggered.connect(lambda: self.lib.refresh_dir())
        add_new_files_action = QAction("&Refresh Directories", menu_bar)
        add_new_files_action.triggered.connect(
            lambda: self.callback_library_needed_check(self.add_new_files_callback)
        )
        add_new_files_action.setShortcut(
            QtCore.QKeyCombination(
                QtCore.Qt.KeyboardModifier(QtCore.Qt.KeyboardModifier.ControlModifier),
                QtCore.Qt.Key.Key_R,
            )
        )
        add_new_files_action.setStatusTip("Ctrl+R")
        # file_menu.addAction(refresh_lib_action)
        file_menu.addAction(add_new_files_action)

        file_menu.addSeparator()

        close_library_action = QAction("&Close Library", menu_bar)
        close_library_action.triggered.connect(lambda: self.close_library())
        file_menu.addAction(close_library_action)

        # Edit Menu ============================================================
        new_tag_action = QAction("New &Tag", menu_bar)
        new_tag_action.triggered.connect(lambda: self.add_tag_action_callback())
        new_tag_action.setShortcut(
            QtCore.QKeyCombination(
                QtCore.Qt.KeyboardModifier(QtCore.Qt.KeyboardModifier.ControlModifier),
                QtCore.Qt.Key.Key_T,
            )
        )
        new_tag_action.setToolTip("Ctrl+T")
        edit_menu.addAction(new_tag_action)

        edit_menu.addSeparator()

        select_all_action = QAction("Select All", menu_bar)
        select_all_action.triggered.connect(self.select_all_action_callback)
        select_all_action.setShortcut(
            QtCore.QKeyCombination(
                QtCore.Qt.KeyboardModifier(QtCore.Qt.KeyboardModifier.ControlModifier),
                QtCore.Qt.Key.Key_A,
            )
        )
        select_all_action.setToolTip("Ctrl+A")
        edit_menu.addAction(select_all_action)

        clear_select_action = QAction("Clear Selection", menu_bar)
        clear_select_action.triggered.connect(self.clear_select_action_callback)
        clear_select_action.setShortcut(QtCore.Qt.Key.Key_Escape)
        clear_select_action.setToolTip("Esc")
        edit_menu.addAction(clear_select_action)

        edit_menu.addSeparator()

        manage_file_extensions_action = QAction("Manage File Extensions", menu_bar)
        manage_file_extensions_action.triggered.connect(
            lambda: self.show_file_extension_modal()
        )
        edit_menu.addAction(manage_file_extensions_action)

        tag_database_action = QAction("Manage Tags", menu_bar)
        tag_database_action.triggered.connect(lambda: self.show_tag_database())
        edit_menu.addAction(tag_database_action)

        check_action = QAction("Open library on start", self)
        check_action.setCheckable(True)
        check_action.setChecked(
            self.settings.value(SettingItems.START_LOAD_LAST, True, type=bool)  # type: ignore
        )
        check_action.triggered.connect(
            lambda checked: self.settings.setValue(
                SettingItems.START_LOAD_LAST, checked
            )
        )
        window_menu.addAction(check_action)

        # Tools Menu ===========================================================
        fix_unlinked_entries_action = QAction("Fix &Unlinked Entries", menu_bar)
        fue_modal = FixUnlinkedEntriesModal(self.lib, self)
        fix_unlinked_entries_action.triggered.connect(lambda: fue_modal.show())
        tools_menu.addAction(fix_unlinked_entries_action)

        fix_dupe_files_action = QAction("Fix Duplicate &Files", menu_bar)
        fdf_modal = FixDupeFilesModal(self.lib, self)
        fix_dupe_files_action.triggered.connect(lambda: fdf_modal.show())
        tools_menu.addAction(fix_dupe_files_action)

        create_collage_action = QAction("Create Collage", menu_bar)
        create_collage_action.triggered.connect(lambda: self.create_collage())
        tools_menu.addAction(create_collage_action)

        # Macros Menu ==========================================================
        self.autofill_action = QAction("Autofill", menu_bar)
        self.autofill_action.triggered.connect(
            lambda: (
                self.run_macros(
                    "autofill", [x[1] for x in self.selected if x[0] == ItemType.ENTRY]
                ),
                self.preview_panel.update_widgets(),
            )
        )
        macros_menu.addAction(self.autofill_action)

        self.sort_fields_action = QAction("&Sort Fields", menu_bar)
        self.sort_fields_action.triggered.connect(
            lambda: (
                self.run_macros(
                    "sort-fields",
                    [x[1] for x in self.selected if x[0] == ItemType.ENTRY],
                ),
                self.preview_panel.update_widgets(),
            )
        )
        self.sort_fields_action.setShortcut(
            QtCore.QKeyCombination(
                QtCore.Qt.KeyboardModifier(QtCore.Qt.KeyboardModifier.AltModifier),
                QtCore.Qt.Key.Key_S,
            )
        )
        self.sort_fields_action.setToolTip("Alt+S")
        macros_menu.addAction(self.sort_fields_action)

        show_libs_list_action = QAction("Show Recent Libraries", menu_bar)
        show_libs_list_action.setCheckable(True)
        show_libs_list_action.setChecked(
            self.settings.value(SettingItems.WINDOW_SHOW_LIBS, True, type=bool)  # type: ignore
        )
        show_libs_list_action.triggered.connect(
            lambda checked: (
                self.settings.setValue(SettingItems.WINDOW_SHOW_LIBS, checked),  # type: ignore
                self.toggle_libs_list(checked),
            )
        )
        window_menu.addAction(show_libs_list_action)

        folders_to_tags_action = QAction("Folders to Tags", menu_bar)
        ftt_modal = FoldersToTagsModal(self.lib, self)
        folders_to_tags_action.triggered.connect(lambda: ftt_modal.show())
        macros_menu.addAction(folders_to_tags_action)

        # Help Menu ==========================================================
        self.repo_action = QAction("Visit GitHub Repository", menu_bar)
        self.repo_action.triggered.connect(
            lambda: webbrowser.open("https://github.com/TagStudioDev/TagStudio")
        )
        help_menu.addAction(self.repo_action)
        self.set_macro_menu_viability()

        menu_bar.addMenu(file_menu)
        menu_bar.addMenu(edit_menu)
        menu_bar.addMenu(tools_menu)
        menu_bar.addMenu(macros_menu)
        menu_bar.addMenu(window_menu)
        menu_bar.addMenu(help_menu)

        self.preview_panel = PreviewPanel(self.lib, self)
        l: QHBoxLayout = self.main_window.splitter
        l.addWidget(self.preview_panel)

        QFontDatabase.addApplicationFont(
            str(Path(__file__).parents[2] / "resources/qt/fonts/Oxanium-Bold.ttf")
        )

        self.thumb_size = 128
        self.max_results = 500
        self.item_thumbs: list[ItemThumb] = []
        self.thumb_renderers: list[ThumbRenderer] = []
        self.collation_thumb_size = math.ceil(self.thumb_size * 2)

        self.init_library_window()

        lib = None
        if self.args.open:
            lib = self.args.open
        elif self.settings.value(SettingItems.START_LOAD_LAST, True, type=bool):
            lib = self.settings.value(SettingItems.LAST_LIBRARY)

            # TODO: Remove this check if the library is no longer saved with files
            if lib and not (Path(lib) / TS_FOLDER_NAME).exists():
                logging.error(
                    f"[QT DRIVER] {TS_FOLDER_NAME} folder in {lib} does not exist."
                )
                self.settings.setValue(SettingItems.LAST_LIBRARY, "")
                lib = None

        if lib:
            self.splash.showMessage(
                f'Opening Library "{lib}"...',
                int(Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignHCenter),
                QColor("#9782ff"),
            )
            self.open_library(Path(lib))

        if self.args.ci:
            # gracefully terminate the app in CI environment
            self.thumb_job_queue.put((self.SIGTERM.emit, []))

        app.exec()

        self.shutdown()

    def init_library_window(self):
        # self._init_landing_page() # Taken care of inside the widget now
        self._init_thumb_grid()

        # TODO: Put this into its own method that copies the font file(s) into memory
        # so the resource isn't being used, then store the specific size variations
        # in a global dict for methods to access for different DPIs.
        # adj_font_size = math.floor(12 * self.main_window.devicePixelRatio())
        # self.ext_font = ImageFont.truetype(os.path.normpath(f'{Path(__file__).parents[2]}/resources/qt/fonts/Oxanium-Bold.ttf'), adj_font_size)

        search_button: QPushButton = self.main_window.searchButton
        search_button.clicked.connect(
            lambda: self.filter_items(self.main_window.searchField.text())
        )
        search_field: QLineEdit = self.main_window.searchField
        search_field.returnPressed.connect(
            lambda: self.filter_items(self.main_window.searchField.text())
        )
        search_type_selector: QComboBox = self.main_window.comboBox_2
        search_type_selector.currentIndexChanged.connect(
            lambda: self.set_search_type(
                SearchMode(search_type_selector.currentIndex())
            )
        )

        back_button: QPushButton = self.main_window.backButton
        back_button.clicked.connect(self.nav_back)
        forward_button: QPushButton = self.main_window.forwardButton
        forward_button.clicked.connect(self.nav_forward)

        # NOTE: Putting this early will result in a white non-responsive
        # window until everything is loaded. Consider adding a splash screen
        # or implementing some clever loading tricks.
        self.main_window.show()
        self.main_window.activateWindow()
        self.main_window.toggle_landing_page(True)

        self.frame_dict = {}
        self.main_window.pagination.index.connect(
            lambda i: (
                self.nav_forward(
                    *self.get_frame_contents(
                        i, self.nav_frames[self.cur_frame_idx].search_text
                    )
                ),
                logging.info(f"emitted {i}"),
            )
        )

        self.nav_frames = []
        self.cur_frame_idx = -1
        self.cur_query = ""
        self.filter_items()
        # self.update_thumbs()

        # self.render_times: list = []
        # self.main_window.setWindowFlag(Qt.FramelessWindowHint)

        # self.main_window.raise_()
        self.splash.finish(self.main_window)
        self.preview_panel.update_widgets()

    def toggle_libs_list(self, value: bool):
        if value:
            self.preview_panel.libs_flow_container.show()
        else:
            self.preview_panel.libs_flow_container.hide()
        self.preview_panel.update()

    def callback_library_needed_check(self, func):
        """Check if loaded library has valid path before executing the button function"""
        if self.lib.library_dir:
            func()

    def handleSIGTERM(self):
        self.shutdown()

    def shutdown(self):
        """Save Library on Application Exit"""
        if self.lib.library_dir:
            self.save_library()
            self.settings.setValue(SettingItems.LAST_LIBRARY, self.lib.library_dir)
            self.settings.sync()
        logging.info("[SHUTDOWN] Ending Thumbnail Threads...")
        for _ in self.thumb_threads:
            self.thumb_job_queue.put(Consumer.MARKER_QUIT)

        # wait for threads to quit
        for thread in self.thumb_threads:
            thread.quit()
            thread.wait()

        QApplication.quit()

    def save_library(self, show_status=True):
        logging.info(f"Saving Library...")
        if show_status:
            self.main_window.statusbar.showMessage(f"Saving Library...")
            start_time = time.time()
        # This might still be able to error, if the selected directory deletes in a race condition
        # or something silly like that. Hence the loop, but if this is considered overkill, thats fair.
        while True:
            try:
                self.lib.save_library_to_disk()
                break
                # If the parent directory got moved, or deleted, prompt user for where to save.
            except FileNotFoundError:
                logging.info(
                    "Library parent directory not found, prompting user to select the directory"
                )
                dir = QFileDialog.getExistingDirectory(
                    None,
                    "Library Location not found, please select location to save Library",
                    "/",
                    QFileDialog.ShowDirsOnly,
                )
                if dir not in (None, ""):
                    self.lib.library_dir = dir
        if show_status:
            end_time = time.time()
            self.main_window.statusbar.showMessage(
                f"Library Saved! ({format_timespan(end_time - start_time)})"
            )

    def close_library(self):
        if self.lib.library_dir:
            logging.info(f"Closing Library...")
            self.main_window.statusbar.showMessage(f"Closing & Saving Library...")
            start_time = time.time()
            self.save_library(show_status=False)
            self.settings.setValue(SettingItems.LAST_LIBRARY, self.lib.library_dir)
            self.settings.sync()

            self.lib.clear_internal_vars()
            title_text = f"{self.base_title}"
            self.main_window.setWindowTitle(title_text)

            self.nav_frames = []
            self.cur_frame_idx = -1
            self.cur_query = ""
            self.selected.clear()
            self.preview_panel.update_widgets()
            self.filter_items()
            self.main_window.toggle_landing_page(True)

            end_time = time.time()
            self.main_window.statusbar.showMessage(
                f"Library Saved and Closed! ({format_timespan(end_time - start_time)})"
            )

    def backup_library(self):
        logging.info(f"Backing Up Library...")
        self.main_window.statusbar.showMessage(f"Saving Library...")
        start_time = time.time()
        fn = self.lib.save_library_backup_to_disk()
        end_time = time.time()
        self.main_window.statusbar.showMessage(
            f'Library Backup Saved at: "{ self.lib.library_dir / TS_FOLDER_NAME / BACKUP_FOLDER_NAME / fn}" ({format_timespan(end_time - start_time)})'
        )

    def add_tag_action_callback(self):
        self.modal = PanelModal(
            BuildTagPanel(self.lib), "New Tag", "Add Tag", has_save=True
        )
        # self.edit_modal.widget.update_display_name.connect(lambda t: self.edit_modal.title_widget.setText(t))
        panel: BuildTagPanel = self.modal.widget
        self.modal.saved.connect(
            lambda: (self.lib.add_tag_to_library(panel.build_tag()), self.modal.hide())
        )
        # panel.tag_updated.connect(lambda tag: self.lib.update_tag(tag))
        self.modal.show()

    def select_all_action_callback(self):
        for item in self.item_thumbs:
            if item.mode and (item.mode, item.item_id) not in self.selected:
                self.selected.append((item.mode, item.item_id))
                item.thumb_button.set_selected(True)

        self.set_macro_menu_viability()
        self.preview_panel.update_widgets()

    def clear_select_action_callback(self):
        self.selected.clear()
        for item in self.item_thumbs:
            item.thumb_button.set_selected(False)

        self.set_macro_menu_viability()
        self.preview_panel.update_widgets()

    def show_tag_database(self):
        self.modal = PanelModal(
            TagDatabasePanel(self.lib), "Library Tags", "Library Tags", has_save=False
        )
        self.modal.show()

    def show_file_extension_modal(self):
        panel = FileExtensionModal(self.lib)
        self.modal = PanelModal(
            panel,
            "File Extensions",
            "File Extensions",
            has_save=True,
        )
        self.modal.saved.connect(lambda: (panel.save(), self.filter_items("")))
        self.modal.show()

    def add_new_files_callback(self):
        """Runs when user initiates adding new files to the Library."""
        # # if self.lib.files_not_in_library:
        # # 	mb = QMessageBox()
        # # 	mb.setText(f'Would you like to refresh the directory before adding {len(self.lib.files_not_in_library)} new files to the library?\nThis will add any additional files that have been moved to the directory since the last refresh.')
        # # 	mb.setWindowTitle('Refresh Library')
        # # 	mb.setIcon(QMessageBox.Icon.Information)
        # # 	mb.setStandardButtons(QMessageBox.StandardButton.No)
        # # 	refresh_button = mb.addButton('Refresh', QMessageBox.ButtonRole.AcceptRole)
        # # 	mb.setDefaultButton(refresh_button)
        # # 	result = mb.exec_()
        # # 	# logging.info(result)
        # # 	if result == 0:
        # # 		self.main_window.statusbar.showMessage(f'Refreshing Library...', 3)
        # # 		self.lib.refresh_dir()
        # # else:
        # pb = QProgressDialog('Scanning Directories for New Files...\nPreparing...', None, 0,0)

        # pb.setFixedSize(432, 112)
        # pb.setWindowFlags(pb.windowFlags() & ~Qt.WindowType.WindowCloseButtonHint)
        # # pb.setLabelText('Scanning Directories...')
        # pb.setWindowTitle('Scanning Directories')
        # pb.setWindowModality(Qt.WindowModality.ApplicationModal)
        # # pb.setMinimum(0)
        # # pb.setMaximum(0)
        # # pb.setValue(0)
        # pb.show()
        # self.main_window.statusbar.showMessage(f'Refreshing Library...', 3)
        # # self.lib.refresh_dir()
        # r = CustomRunnable(lambda: self.runnable(pb))
        # logging.info(f'Main: {QThread.currentThread()}')
        # r.done.connect(lambda: (pb.hide(), pb.deleteLater(), self.add_new_files_runnable()))
        # QThreadPool.globalInstance().start(r)
        # # r.run()

        # # new_ids: list[int] = self.lib.add_new_files_as_entries()
        # # # logging.info(f'{INFO} Running configured Macros on {len(new_ids)} new Entries...')
        # # # self.main_window.statusbar.showMessage(f'Running configured Macros on {len(new_ids)} new Entries...', 3)
        # # # for id in new_ids:
        # # # 	self.run_macro('autofill', id)

        # # self.main_window.statusbar.showMessage('', 3)
        # # self.filter_entries('')

        iterator = FunctionIterator(self.lib.refresh_dir)
        pw = ProgressWidget(
            window_title="Refreshing Directories",
            label_text="Scanning Directories for New Files...\nPreparing...",
            cancel_button_text=None,
            minimum=0,
            maximum=0,
        )
        pw.show()
        iterator.value.connect(lambda x: pw.update_progress(x + 1))
        iterator.value.connect(
            lambda x: pw.update_label(
                f'Scanning Directories for New Files...\n{x + 1} File{"s" if x + 1 != 1 else ""} Searched, {len(self.lib.files_not_in_library)} New Files Found'
            )
        )
        r = CustomRunnable(lambda: iterator.run())
        # r.done.connect(lambda: (pw.hide(), pw.deleteLater(), self.filter_items('')))
        # vvv This one runs the macros when adding new files to the library.
        r.done.connect(
            lambda: (pw.hide(), pw.deleteLater(), self.add_new_files_runnable())
        )
        QThreadPool.globalInstance().start(r)

    # def runnable(self, pb:QProgressDialog):
    # 	for i in self.lib.refresh_dir():
    # 		pb.setLabelText(f'Scanning Directories for New Files...\n{i} File{"s" if i != 1 else ""} Searched, {len(self.lib.files_not_in_library)} New Files Found')

    def add_new_files_runnable(self):
        """
        Threaded method that adds any known new files to the library and
        initiates running default macros on them.
        """
        # logging.info(f'Start ANF: {QThread.currentThread()}')
        new_ids: list[int] = self.lib.add_new_files_as_entries()
        # pb = QProgressDialog(f'Running Configured Macros on 1/{len(new_ids)} New Entries', None, 0,len(new_ids))
        # pb.setFixedSize(432, 112)
        # pb.setWindowFlags(pb.windowFlags() & ~Qt.WindowType.WindowCloseButtonHint)
        # pb.setWindowTitle('Running Macros')
        # pb.setWindowModality(Qt.WindowModality.ApplicationModal)
        # pb.show()

        # r = CustomRunnable(lambda: self.new_file_macros_runnable(pb, new_ids))
        # r.done.connect(lambda: (pb.hide(), pb.deleteLater(), self.filter_items('')))
        # r.run()
        # # QThreadPool.globalInstance().start(r)

        # # logging.info(f'{INFO} Running configured Macros on {len(new_ids)} new Entries...')
        # # self.main_window.statusbar.showMessage(f'Running configured Macros on {len(new_ids)} new Entries...', 3)

        # # pb.hide()

        iterator = FunctionIterator(lambda: self.new_file_macros_runnable(new_ids))
        pw = ProgressWidget(
            window_title="Running Macros on New Entries",
            label_text=f"Running Configured Macros on 1/{len(new_ids)} New Entries",
            cancel_button_text=None,
            minimum=0,
            maximum=0,
        )
        pw.show()
        iterator.value.connect(lambda x: pw.update_progress(x + 1))
        iterator.value.connect(
            lambda x: pw.update_label(
                f"Running Configured Macros on {x + 1}/{len(new_ids)} New Entries"
            )
        )
        r = CustomRunnable(lambda: iterator.run())
        r.done.connect(lambda: (pw.hide(), pw.deleteLater(), self.filter_items("")))
        QThreadPool.globalInstance().start(r)

    def new_file_macros_runnable(self, new_ids):
        """Threaded method that runs macros on a set of Entry IDs."""
        # sleep(1)
        # logging.info(f'ANFR: {QThread.currentThread()}')
        # for i, id in enumerate(new_ids):
        # 	# pb.setValue(i)
        # 	# pb.setLabelText(f'Running Configured Macros on {i}/{len(new_ids)} New Entries')
        # 	# self.run_macro('autofill', id)

        # NOTE: I don't know. I don't know why it needs this. The whole program
        # falls apart if this method doesn't run, and it DOESN'T DO ANYTHING
        yield 0

        # self.main_window.statusbar.showMessage('', 3)

        # sleep(5)
        # pb.deleteLater()

    def run_macros(self, name: str, entry_ids: list[int]):
        """Runs a specific Macro on a group of given entry_ids."""
        for id in entry_ids:
            self.run_macro(name, id)

    def run_macro(self, name: str, entry_id: int):
        """Runs a specific Macro on an Entry given a Macro name."""
        entry = self.lib.get_entry(entry_id)
        path = self.lib.library_dir / entry.path / entry.filename
        source = entry.path.parts[0]
        if name == "sidecar":
            self.lib.add_generic_data_to_entry(
                self.core.get_gdl_sidecar(path, source), entry_id
            )
        elif name == "autofill":
            self.run_macro("sidecar", entry_id)
            self.run_macro("build-url", entry_id)
            self.run_macro("match", entry_id)
            self.run_macro("clean-url", entry_id)
            self.run_macro("sort-fields", entry_id)
        elif name == "build-url":
            data = {"source": self.core.build_url(entry_id, source)}
            self.lib.add_generic_data_to_entry(data, entry_id)
        elif name == "sort-fields":
            order: list[int] = (
                [0]
                + [1, 2]
                + [9, 17, 18, 19, 20]
                + [8, 7, 6]
                + [4]
                + [3, 21]
                + [10, 14, 11, 12, 13, 22]
                + [5]
            )
            self.lib.sort_fields(entry_id, order)
        elif name == "match":
            self.core.match_conditions(entry_id)
        # elif name == 'scrape':
        # 	self.core.scrape(entry_id)
        elif name == "clean-url":
            # entry = self.lib.get_entry_from_index(entry_id)
            if entry.fields:
                for i, field in enumerate(entry.fields, start=0):
                    if self.lib.get_field_attr(field, "type") == "text_line":
                        self.lib.update_entry_field(
                            entry_id=entry_id,
                            field_index=i,
                            content=strip_web_protocol(
                                self.lib.get_field_attr(field, "content")
                            ),
                            mode="replace",
                        )

    def mouse_navigation(self, event: QMouseEvent):
        # print(event.button())
        if event.button() == Qt.MouseButton.ForwardButton:
            self.nav_forward()
        elif event.button() == Qt.MouseButton.BackButton:
            self.nav_back()

    def nav_forward(
        self,
        frame_content: Optional[list[tuple[ItemType, int]]] = None,
        page_index: int = 0,
        page_count: int = 0,
    ):
        """Navigates a step further into the navigation stack."""
        logging.info(
            f"Calling NavForward with Content:{False if not frame_content else frame_content[0]}, Index:{page_index}, PageCount:{page_count}"
        )

        # Ex. User visits | A ->[B]     |
        #                 | A    B ->[C]|
        #                 | A   [B]<- C |
        #                 |[A]<- B    C |  Previous routes still exist
        #                 | A ->[D]     |  Stack is cut from [:A] on new route

        # Moving forward (w/ or wo/ new content) in the middle of the stack
        original_pos = self.cur_frame_idx
        sb: QScrollArea = self.main_window.scrollArea
        sb_pos = sb.verticalScrollBar().value()
        search_text = self.main_window.searchField.text()

        trimmed = False
        if len(self.nav_frames) > self.cur_frame_idx + 1:
            if frame_content is not None:
                # Trim the nav stack if user is taking a new route.
                self.nav_frames = self.nav_frames[: self.cur_frame_idx + 1]
                if self.nav_frames and not self.nav_frames[self.cur_frame_idx].contents:
                    self.nav_frames.pop()
                    trimmed = True
                self.nav_frames.append(
                    NavigationState(
                        frame_content, 0, page_index, page_count, search_text
                    )
                )
                # logging.info(f'Saving Text: {search_text}')
            # Update the last frame's scroll_pos
            self.nav_frames[self.cur_frame_idx].scrollbar_pos = sb_pos
            self.cur_frame_idx += 1 if not trimmed else 0
        # Moving forward at the end of the stack with new content
        elif frame_content is not None:
            # If the current page is empty, don't include it in the new stack.
            if self.nav_frames and not self.nav_frames[self.cur_frame_idx].contents:
                self.nav_frames.pop()
                trimmed = True
            self.nav_frames.append(
                NavigationState(frame_content, 0, page_index, page_count, search_text)
            )
            # logging.info(f'Saving Text: {search_text}')
            self.nav_frames[self.cur_frame_idx].scrollbar_pos = sb_pos
            self.cur_frame_idx += 1 if not trimmed else 0

        # if self.nav_stack[self.cur_page_idx].contents:
        if (self.cur_frame_idx != original_pos) or (frame_content is not None):
            self.update_thumbs()
            sb.verticalScrollBar().setValue(
                self.nav_frames[self.cur_frame_idx].scrollbar_pos
            )
            self.main_window.searchField.setText(
                self.nav_frames[self.cur_frame_idx].search_text
            )
            self.main_window.pagination.update_buttons(
                self.nav_frames[self.cur_frame_idx].page_count,
                self.nav_frames[self.cur_frame_idx].page_index,
                emit=False,
            )
            # logging.info(f'Setting Text: {self.nav_stack[self.cur_page_idx].search_text}')
        # else:
        # 	self.nav_stack.pop()
        # 	self.cur_page_idx -= 1
        # 	self.update_thumbs()
        # 	sb.verticalScrollBar().setValue(self.nav_stack[self.cur_page_idx].scrollbar_pos)

        # logging.info(f'Forward: {[len(x.contents) for x in self.nav_stack]}, Index {self.cur_page_idx}, SB {self.nav_stack[self.cur_page_idx].scrollbar_pos}')

    def nav_back(self):
        """Navigates a step backwards in the navigation stack."""

        original_pos = self.cur_frame_idx
        sb: QScrollArea = self.main_window.scrollArea
        sb_pos = sb.verticalScrollBar().value()

        if self.cur_frame_idx > 0:
            self.nav_frames[self.cur_frame_idx].scrollbar_pos = sb_pos
            self.cur_frame_idx -= 1
            if self.cur_frame_idx != original_pos:
                self.update_thumbs()
                sb.verticalScrollBar().setValue(
                    self.nav_frames[self.cur_frame_idx].scrollbar_pos
                )
                self.main_window.searchField.setText(
                    self.nav_frames[self.cur_frame_idx].search_text
                )
                self.main_window.pagination.update_buttons(
                    self.nav_frames[self.cur_frame_idx].page_count,
                    self.nav_frames[self.cur_frame_idx].page_index,
                    emit=False,
                )
                # logging.info(f'Setting Text: {self.nav_stack[self.cur_page_idx].search_text}')
        # logging.info(f'Back: {[len(x.contents) for x in self.nav_stack]}, Index {self.cur_page_idx}, SB {self.nav_stack[self.cur_page_idx].scrollbar_pos}')

    def refresh_frame(
        self,
        frame_content: list[tuple[ItemType, int]],
        page_index: int = 0,
        page_count: int = 0,
    ):
        """
        Refreshes the current navigation contents without altering the
        navigation stack order.
        """
        if self.nav_frames:
            self.nav_frames[self.cur_frame_idx] = NavigationState(
                frame_content,
                0,
                self.nav_frames[self.cur_frame_idx].page_index,
                self.nav_frames[self.cur_frame_idx].page_count,
                self.main_window.searchField.text(),
            )
        else:
            self.nav_forward(frame_content, page_index, page_count)
        self.update_thumbs()
        # logging.info(f'Refresh: {[len(x.contents) for x in self.nav_stack]}, Index {self.cur_page_idx}')

    @typing.no_type_check
    def purge_item_from_navigation(self, type: ItemType, id: int):
        # logging.info(self.nav_frames)
        # TODO - types here are ambiguous
        for i, frame in enumerate(self.nav_frames, start=0):
            while (type, id) in frame.contents:
                logging.info(f"Removing {id} from nav stack frame {i}")
                frame.contents.remove((type, id))

        for i, key in enumerate(self.frame_dict.keys(), start=0):
            for frame in self.frame_dict[key]:
                while (type, id) in frame:
                    logging.info(f"Removing {id} from frame dict item {i}")
                    frame.remove((type, id))

        while (type, id) in self.selected:
            logging.info(f"Removing {id} from frame selected")
            self.selected.remove((type, id))

    def _init_thumb_grid(self):
        # logging.info('Initializing Thumbnail Grid...')
        layout = FlowLayout()
        layout.setGridEfficiency(True)
        # layout.setContentsMargins(0,0,0,0)
        layout.setSpacing(min(self.thumb_size // 10, 12))
        # layout = QHBoxLayout()
        # layout.setSizeConstraint(QLayout.SizeConstraint.SetMaximumSize)
        # layout = QListView()
        # layout.setViewMode(QListView.ViewMode.IconMode)

        col_size = 28
        for i in range(0, self.max_results):
            item_thumb = ItemThumb(
                None, self.lib, self.preview_panel, (self.thumb_size, self.thumb_size)
            )
            layout.addWidget(item_thumb)
            self.item_thumbs.append(item_thumb)

        self.flow_container: QWidget = QWidget()
        self.flow_container.setObjectName("flowContainer")
        self.flow_container.setLayout(layout)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        sa: QScrollArea = self.main_window.scrollArea
        sa.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        sa.setWidgetResizable(True)
        sa.setWidget(self.flow_container)

    def select_item(self, type: ItemType, id: int, append: bool, bridge: bool):
        """Selects one or more items in the Thumbnail Grid."""
        if append:
            # self.selected.append((thumb_index, page_index))
            if ((type, id)) not in self.selected:
                self.selected.append((type, id))
                for it in self.item_thumbs:
                    if it.mode == type and it.item_id == id:
                        it.thumb_button.set_selected(True)
            else:
                self.selected.remove((type, id))
                for it in self.item_thumbs:
                    if it.mode == type and it.item_id == id:
                        it.thumb_button.set_selected(False)
            # self.item_thumbs[thumb_index].thumb_button.set_selected(True)

        elif bridge and self.selected:
            logging.info(f"Last Selected: {self.selected[-1]}")
            contents = self.nav_frames[self.cur_frame_idx].contents
            last_index = self.nav_frames[self.cur_frame_idx].contents.index(
                self.selected[-1]
            )
            current_index = self.nav_frames[self.cur_frame_idx].contents.index(
                (type, id)
            )
            index_range: list = contents[
                min(last_index, current_index) : max(last_index, current_index) + 1
            ]
            # Preserve bridge direction for correct appending order.
            if last_index < current_index:
                index_range.reverse()

            # logging.info(f'Current Frame Contents: {len(self.nav_frames[self.cur_frame_idx].contents)}')
            # logging.info(f'Last Selected Index: {last_index}')
            # logging.info(f'Current Selected Index: {current_index}')
            # logging.info(f'Index Range: {index_range}')

            for c_type, c_id in index_range:
                for it in self.item_thumbs:
                    if it.mode == c_type and it.item_id == c_id:
                        it.thumb_button.set_selected(True)
                        if ((c_type, c_id)) not in self.selected:
                            self.selected.append((c_type, c_id))
        else:
            # for i in self.selected:
            # 	if i[1] == self.cur_frame_idx:
            # 		self.item_thumbs[i[0]].thumb_button.set_selected(False)
            self.selected.clear()
            # self.selected.append((thumb_index, page_index))
            self.selected.append((type, id))
            # self.item_thumbs[thumb_index].thumb_button.set_selected(True)
            for it in self.item_thumbs:
                if it.mode == type and it.item_id == id:
                    it.thumb_button.set_selected(True)
                else:
                    it.thumb_button.set_selected(False)

        # NOTE: By using the preview panel's "set_tags_updated_slot" method,
        # only the last of multiple identical item selections are connected.
        # If attaching the slot to multiple duplicate selections is needed,
        # just bypass the method and manually disconnect and connect the slots.
        if len(self.selected) == 1:
            for it in self.item_thumbs:
                if it.mode == type and it.item_id == id:
                    self.preview_panel.set_tags_updated_slot(it.update_badges)

        self.set_macro_menu_viability()
        self.preview_panel.update_widgets()

    def set_macro_menu_viability(self):
        if len([x[1] for x in self.selected if x[0] == ItemType.ENTRY]) == 0:
            self.autofill_action.setDisabled(True)
            self.sort_fields_action.setDisabled(True)
        else:
            self.autofill_action.setDisabled(False)
            self.sort_fields_action.setDisabled(False)

    def update_thumbs(self):
        """Updates search thumbnails."""
        # start_time = time.time()
        # logging.info(f'Current Page: {self.cur_page_idx}, Stack Length:{len(self.nav_stack)}')
        with self.thumb_job_queue.mutex:
            # Cancels all thumb jobs waiting to be started
            self.thumb_job_queue.queue.clear()
            self.thumb_job_queue.all_tasks_done.notify_all()
            self.thumb_job_queue.not_full.notify_all()
            # Stops in-progress jobs from finishing
            ItemThumb.update_cutoff = time.time()

        ratio: float = self.main_window.devicePixelRatio()
        base_size: tuple[int, int] = (self.thumb_size, self.thumb_size)

        for i, item_thumb in enumerate(self.item_thumbs, start=0):
            if i < len(self.nav_frames[self.cur_frame_idx].contents):
                # Set new item type modes
                # logging.info(f'[UPDATE] Setting Mode To: {self.nav_stack[self.cur_page_idx].contents[i][0]}')
                item_thumb.set_mode(self.nav_frames[self.cur_frame_idx].contents[i][0])
                item_thumb.ignore_size = False
                # logging.info(f'[UPDATE] Set Mode To: {item.mode}')
                # Set thumbnails to loading (will always finish if rendering)
                self.thumb_job_queue.put(
                    (
                        item_thumb.renderer.render,
                        (sys.float_info.max, "", base_size, ratio, True, True),
                    )
                )
                # # Restore Selected Borders
                # if (item_thumb.mode, item_thumb.item_id) in self.selected:
                # 	item_thumb.thumb_button.set_selected(True)
                # else:
                # 	item_thumb.thumb_button.set_selected(False)
            else:
                item_thumb.ignore_size = True
                item_thumb.set_mode(None)
                item_thumb.set_item_id(-1)
                item_thumb.thumb_button.set_selected(False)

        # scrollbar: QScrollArea = self.main_window.scrollArea
        # scrollbar.verticalScrollBar().setValue(scrollbar_pos)
        self.flow_container.layout().update()
        self.main_window.update()

        for i, item_thumb in enumerate(self.item_thumbs, start=0):
            if i < len(self.nav_frames[self.cur_frame_idx].contents):
                filepath = ""
                if self.nav_frames[self.cur_frame_idx].contents[i][0] == ItemType.ENTRY:
                    entry = self.lib.get_entry(
                        self.nav_frames[self.cur_frame_idx].contents[i][1]
                    )
                    filepath = self.lib.library_dir / entry.path / entry.filename

                    item_thumb.set_item_id(entry.id)
                    item_thumb.assign_archived(entry.has_tag(self.lib, TAG_ARCHIVED))
                    item_thumb.assign_favorite(entry.has_tag(self.lib, TAG_FAVORITE))
                    # ctrl_down = True if QGuiApplication.keyboardModifiers() else False
                    # TODO: Change how this works. The click function
                    # for collations a few lines down should NOT be allowed during modifier keys.
                    item_thumb.update_clickable(
                        clickable=(
                            lambda checked=False, entry=entry: self.select_item(
                                ItemType.ENTRY,
                                entry.id,
                                append=True
                                if QGuiApplication.keyboardModifiers()
                                == Qt.KeyboardModifier.ControlModifier
                                else False,
                                bridge=True
                                if QGuiApplication.keyboardModifiers()
                                == Qt.KeyboardModifier.ShiftModifier
                                else False,
                            )
                        )
                    )
                    # item_thumb.update_clickable(clickable=(
                    # 	lambda checked=False, filepath=filepath, entry=entry,
                    # 		   item_t=item_thumb, i=i, page=self.cur_frame_idx: (
                    # 		self.preview_panel.update_widgets(entry),
                    # 		self.select_item(ItemType.ENTRY, entry.id,
                    # 	append=True if QGuiApplication.keyboardModifiers() == Qt.KeyboardModifier.ControlModifier else False,
                    # 	bridge=True if QGuiApplication.keyboardModifiers() == Qt.KeyboardModifier.ShiftModifier else False))))
                    # item.dumpObjectTree()
                elif (
                    self.nav_frames[self.cur_frame_idx].contents[i][0]
                    == ItemType.COLLATION
                ):
                    collation = self.lib.get_collation(
                        self.nav_frames[self.cur_frame_idx].contents[i][1]
                    )
                    cover_id = (
                        collation.cover_id
                        if collation.cover_id >= 0
                        else collation.e_ids_and_pages[0][0]
                    )
                    cover_e = self.lib.get_entry(cover_id)
                    filepath = self.lib.library_dir / cover_e.path / cover_e.filename
                    item_thumb.set_count(str(len(collation.e_ids_and_pages)))
                    item_thumb.update_clickable(
                        clickable=(
                            lambda checked=False,
                            filepath=filepath,
                            entry=cover_e,
                            collation=collation: (
                                self.expand_collation(collation.e_ids_and_pages)
                            )
                        )
                    )
                # item.setHidden(False)

                # Restore Selected Borders
                if (item_thumb.mode, item_thumb.item_id) in self.selected:
                    item_thumb.thumb_button.set_selected(True)
                else:
                    item_thumb.thumb_button.set_selected(False)

                self.thumb_job_queue.put(
                    (
                        item_thumb.renderer.render,
                        (time.time(), filepath, base_size, ratio, False, True),
                    )
                )
            else:
                # item.setHidden(True)
                pass
                # update_widget_clickable(widget=item.bg_button, clickable=())
                # self.thumb_job_queue.put(
                # 	(item.renderer.render, ('', base_size, ratio, False)))

        # end_time = time.time()
        # logging.info(
        # 	f'[MAIN] Elements thumbs updated in {(end_time - start_time):.3f} seconds')

    def update_badges(self):
        for i, item_thumb in enumerate(self.item_thumbs, start=0):
            item_thumb.update_badges()

    def expand_collation(self, collation_entries: list[tuple[int, int]]):
        self.nav_forward([(ItemType.ENTRY, x[0]) for x in collation_entries])
        # self.update_thumbs()

    def get_frame_contents(self, index=0, query: str = ""):
        return (
            [] if not self.frame_dict[query] else self.frame_dict[query][index],
            index,
            len(self.frame_dict[query]),
        )

    def filter_items(self, query: str = ""):
        if self.lib:
            # logging.info('Filtering...')
            self.main_window.statusbar.showMessage(
                f'Searching Library for "{query}"...'
            )
            self.main_window.statusbar.repaint()
            start_time = time.time()

            # self.filtered_items = self.lib.search_library(query)
            # 73601 Entries at 500 size should be 246
            all_items = self.lib.search_library(query, search_mode=self.search_mode)
            frames: list[list[tuple[ItemType, int]]] = []
            frame_count = math.ceil(len(all_items) / self.max_results)
            for i in range(0, frame_count):
                frames.append(
                    all_items[
                        min(len(all_items) - 1, (i) * self.max_results) : min(
                            len(all_items), (i + 1) * self.max_results
                        )
                    ]
                )
            for i, f in enumerate(frames):
                logging.info(f"Query:{query}, Frame: {i},  Length: {len(f)}")
            self.frame_dict[query] = frames
            # self.frame_dict[query] = [all_items]

            if self.cur_query == query:
                # self.refresh_frame(self.lib.search_library(query))
                # NOTE: Trying to refresh instead of navigating forward here
                # now creates a bug when the page counts differ on refresh.
                # If refreshing is absolutely desired, see how to update
                # page counts where they need to be updated.
                self.nav_forward(*self.get_frame_contents(0, query))
            else:
                # self.nav_forward(self.lib.search_library(query))
                self.nav_forward(*self.get_frame_contents(0, query))
            self.cur_query = query

            end_time = time.time()
            if query:
                self.main_window.statusbar.showMessage(
                    f'{len(all_items)} Results Found for "{query}" ({format_timespan(end_time - start_time)})'
                )
            else:
                self.main_window.statusbar.showMessage(
                    f"{len(all_items)} Results ({format_timespan(end_time - start_time)})"
                )
            # logging.info(f'Done Filtering! ({(end_time - start_time):.3f}) seconds')

            # self.update_thumbs()

    def set_search_type(self, mode=SearchMode.AND):
        self.search_mode = mode
        self.filter_items(self.main_window.searchField.text())

    def remove_recent_library(self, item_key: str):
        self.settings.beginGroup(SettingItems.LIBS_LIST)
        self.settings.remove(item_key)
        self.settings.endGroup()
        self.settings.sync()

    @typing.no_type_check
    def update_libs_list(self, path: Path):
        """add library to list in SettingItems.LIBS_LIST"""
        ITEMS_LIMIT = 5
        path = Path(path)

        self.settings.beginGroup(SettingItems.LIBS_LIST)

        all_libs = {str(time.time()): str(path)}

        for item_key in self.settings.allKeys():
            item_path = self.settings.value(item_key)
            if Path(item_path) != path:
                all_libs[item_key] = item_path

        # sort items, most recent first
        all_libs = sorted(all_libs.items(), key=lambda item: item[0], reverse=True)

        # remove previously saved items
        self.settings.clear()

        for item_key, item_value in all_libs[:ITEMS_LIMIT]:
            self.settings.setValue(item_key, item_value)

        self.settings.endGroup()
        self.settings.sync()

    def open_library(self, path: Path):
        """Opens a TagStudio library."""
        open_message: str = f'Opening Library "{str(path)}"...'
        self.main_window.landing_widget.set_status_label(open_message)
        self.main_window.statusbar.showMessage(open_message, 3)
        self.main_window.repaint()

        if self.lib.library_dir:
            self.save_library()
            self.lib.clear_internal_vars()

        return_code = self.lib.open_library(path)
        if return_code == 1:
            pass
        else:
            logging.info(
                f"{ERROR} No existing TagStudio library found at '{path}'. Creating one."
            )
            print(f"Library Creation Return Code: {self.lib.create_library(path)}")
            self.add_new_files_callback()

        self.update_libs_list(path)
        title_text = f"{self.base_title} - Library '{self.lib.library_dir}'"
        self.main_window.setWindowTitle(title_text)

        self.nav_frames = []
        self.cur_frame_idx = -1
        self.cur_query = ""
        self.selected.clear()
        self.preview_panel.update_widgets()
        self.filter_items()
        self.main_window.toggle_landing_page(False)

    def create_collage(self) -> None:
        """Generates and saves an image collage based on Library Entries."""

        run: bool = True
        keep_aspect: bool = False
        data_only_mode: bool = False
        data_tint_mode: bool = False

        self.main_window.statusbar.showMessage(f"Creating Library Collage...")
        self.collage_start_time = time.time()

        # mode:int = self.scr_choose_option(subtitle='Choose Collage Mode(s)',
        # 	choices=[
        # 	('Normal','Creates a standard square image collage made up of Library media files.'),
        # 	('Data Tint','Tints the collage with a color representing data about the Library Entries/files.'),
        # 	('Data Only','Ignores media files entirely and only outputs a collage of Library Entry/file data.'),
        # 	('Normal & Data Only','Creates both Normal and Data Only collages.'),
        # 	], prompt='', required=True)
        mode = 0

        if mode == 1:
            data_tint_mode = True

        if mode == 2:
            data_only_mode = True

        if mode in [0, 1, 3]:
            # keep_aspect = self.scr_choose_option(
            # 	subtitle='Choose Aspect Ratio Option',
            # 	choices=[
            # 	('Stretch to Fill','Stretches the media file to fill the entire collage square.'),
            # 	('Keep Aspect Ratio','Keeps the original media file\'s aspect ratio, filling the rest of the square with black bars.')
            # 	], prompt='', required=True)
            keep_aspect = False

        if mode in [1, 2, 3]:
            # TODO: Choose data visualization options here.
            pass

        full_thumb_size: int = 1

        if mode in [0, 1, 3]:
            # full_thumb_size = self.scr_choose_option(
            # 	subtitle='Choose Thumbnail Size',
            # 	choices=[
            # 	('Tiny (32px)',''),
            # 	('Small (64px)',''),
            # 	('Medium (128px)',''),
            # 	('Large (256px)',''),
            # 	('Extra Large (512px)','')
            # 	], prompt='', required=True)
            full_thumb_size = 0

        thumb_size: int = (
            32
            if (full_thumb_size == 0)
            else 64
            if (full_thumb_size == 1)
            else 128
            if (full_thumb_size == 2)
            else 256
            if (full_thumb_size == 3)
            else 512
            if (full_thumb_size == 4)
            else 32
        )
        thumb_size = 16

        # if len(com) > 1 and com[1] == 'keep-aspect':
        # 	keep_aspect = True
        # elif len(com) > 1 and com[1] == 'data-only':
        # 	data_only_mode = True
        # elif len(com) > 1 and com[1] == 'data-tint':
        # 	data_tint_mode = True
        grid_size = math.ceil(math.sqrt(len(self.lib.entries))) ** 2
        grid_len = math.floor(math.sqrt(grid_size))
        thumb_size = thumb_size if not data_only_mode else 1
        img_size = thumb_size * grid_len

        logging.info(
            f"Creating collage for {len(self.lib.entries)} Entries.\nGrid Size: {grid_size} ({grid_len}x{grid_len})\nIndividual Picture Size: ({thumb_size}x{thumb_size})"
        )
        if keep_aspect:
            logging.info("Keeping original aspect ratios.")
        if data_only_mode:
            logging.info("Visualizing Entry Data")

        if not data_only_mode:
            time.sleep(5)

        self.collage = Image.new("RGB", (img_size, img_size))
        i = 0
        self.completed = 0
        for x in range(0, grid_len):
            for y in range(0, grid_len):
                if i < len(self.lib.entries) and run:
                    # if i < 5 and run:

                    entry_id = self.lib.entries[i].id
                    renderer = CollageIconRenderer(self.lib)
                    renderer.rendered.connect(
                        lambda image, x=x, y=y: self.collage.paste(
                            image, (y * thumb_size, x * thumb_size)
                        )
                    )
                    renderer.done.connect(lambda: self.try_save_collage(True))
                    self.thumb_job_queue.put(
                        (
                            renderer.render,
                            (
                                entry_id,
                                (thumb_size, thumb_size),
                                data_tint_mode,
                                data_only_mode,
                                keep_aspect,
                            ),
                        )
                    )
                i = i + 1

    def try_save_collage(self, increment_progress: bool):
        if increment_progress:
            self.completed += 1
        # logging.info(f'threshold:{len(self.lib.entries}, completed:{self.completed}')
        if self.completed == len(self.lib.entries):
            filename = (
                self.lib.library_dir
                / TS_FOLDER_NAME
                / COLLAGE_FOLDER_NAME
                / f'collage_{dt.utcnow().strftime("%F_%T").replace(":", "")}.png'
            )
            self.collage.save(filename)
            self.collage = None

            end_time = time.time()
            self.main_window.statusbar.showMessage(
                f'Collage Saved at "{filename}" ({format_timespan(end_time - self.collage_start_time)})'
            )
            logging.info(
                f'Collage Saved at "{filename}" ({format_timespan(end_time - self.collage_start_time)})'
            )



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from PySide6.QtCore import Signal, Qt
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QComboBox,
)

from src.core.library import Library


class AddFieldModal(QWidget):
    done = Signal(int)

    def __init__(self, library: "Library"):
        # [Done]
        # - OR -
        # [Cancel] [Save]
        super().__init__()
        self.is_connected = False
        self.lib = library
        self.setWindowTitle(f"Add Field")
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setMinimumSize(400, 300)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 6, 6, 6)

        self.title_widget = QLabel()
        self.title_widget.setObjectName("fieldTitle")
        self.title_widget.setWordWrap(True)
        self.title_widget.setStyleSheet(
            # 'background:blue;'
            # 'text-align:center;'
            "font-weight:bold;" "font-size:14px;" "padding-top: 6px" ""
        )
        self.title_widget.setText("Add Field")
        self.title_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.combo_box = QComboBox()
        self.combo_box.setEditable(False)
        # self.combo_box.setMaxVisibleItems(5)
        self.combo_box.setStyleSheet("combobox-popup:0;")
        self.combo_box.view().setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAsNeeded
        )
        for df in self.lib.default_fields:
            self.combo_box.addItem(
                f'{df["name"]} ({df["type"].replace("_", " ").title()})'
            )

        self.button_container = QWidget()
        self.button_layout = QHBoxLayout(self.button_container)
        self.button_layout.setContentsMargins(6, 6, 6, 6)
        self.button_layout.addStretch(1)

        # self.cancel_button = QPushButton()
        # self.cancel_button.setText('Cancel')

        self.cancel_button = QPushButton()
        self.cancel_button.setText("Cancel")
        self.cancel_button.clicked.connect(self.hide)
        # self.cancel_button.clicked.connect(widget.reset)
        self.button_layout.addWidget(self.cancel_button)

        self.save_button = QPushButton()
        self.save_button.setText("Add")
        # self.save_button.setAutoDefault(True)
        self.save_button.setDefault(True)
        self.save_button.clicked.connect(self.hide)
        self.save_button.clicked.connect(
            lambda: self.done.emit(self.combo_box.currentIndex())
        )
        # self.save_button.clicked.connect(lambda: save_callback(widget.get_content()))
        self.button_layout.addWidget(self.save_button)

        # self.returnPressed.connect(lambda: self.done.emit(self.combo_box.currentIndex()))

        # self.done.connect(lambda x: callback(x))

        self.root_layout.addWidget(self.title_widget)
        self.root_layout.addWidget(self.combo_box)
        # self.root_layout.setStretch(1,2)
        self.root_layout.addStretch(1)
        self.root_layout.addWidget(self.button_container)



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

import typing

from PySide6.QtCore import QObject, Signal, QThreadPool

from src.core.library import Library
from src.qt.helpers.function_iterator import FunctionIterator
from src.qt.helpers.custom_runnable import CustomRunnable
from src.qt.widgets.progress import ProgressWidget

# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver


class MergeDuplicateEntries(QObject):
    done = Signal()

    def __init__(self, library: "Library", driver: "QtDriver"):
        super().__init__()
        self.lib = library
        self.driver = driver

    def merge_entries(self):
        iterator = FunctionIterator(self.lib.merge_dupe_entries)

        pw = ProgressWidget(
            window_title="Merging Duplicate Entries",
            label_text="",
            cancel_button_text=None,
            minimum=0,
            maximum=len(self.lib.dupe_entries),
        )
        pw.show()

        iterator.value.connect(lambda x: pw.update_progress(x))
        iterator.value.connect(
            lambda: (pw.update_label("Merging Duplicate Entries..."))
        )

        r = CustomRunnable(lambda: iterator.run())
        r.done.connect(lambda: (pw.hide(), pw.deleteLater(), self.done.emit()))
        QThreadPool.globalInstance().start(r)



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from PySide6.QtCore import Signal, QRunnable, QObject


class CustomRunnable(QRunnable, QObject):
    done = Signal()

    def __init__(self, function) -> None:
        QRunnable.__init__(self)
        QObject.__init__(self)
        self.function = function

    def run(self):
        self.function()
        self.done.emit()



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio
from typing import Callable

from PySide6.QtWidgets import QVBoxLayout, QLineEdit

from src.qt.widgets.panel import PanelWidget


class EditTextLine(PanelWidget):
    def __init__(self, text):
        super().__init__()
        # self.setLayout()
        self.setMinimumWidth(480)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 0, 6, 0)
        self.text = text
        self.text_edit = QLineEdit()
        self.text_edit.setText(text)
        self.root_layout.addWidget(self.text_edit)

    def get_content(self) -> str:
        return self.text_edit.text()

    def reset(self):
        self.text_edit.setText(self.text)

    def add_callback(self, callback: Callable, event: str = "returnPressed"):
        if event == "returnPressed":
            self.text_edit.returnPressed.connect(callback)
        else:
            raise ValueError(f"unknown event type: {event}")



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio
import logging
from types import FunctionType
from typing import Callable

from PySide6.QtCore import Signal, Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QPushButton


class PanelModal(QWidget):
    saved = Signal()

    # TODO: Separate callbacks from the buttons you want, and just generally
    # figure out what you want from this.
    def __init__(
        self,
        widget: "PanelWidget",
        title: str,
        window_title: str,
        done_callback: Callable = None,
        #  cancel_callback:FunctionType=None,
        save_callback: Callable = None,
        has_save: bool = False,
    ):
        # [Done]
        # - OR -
        # [Cancel] [Save]
        super().__init__()
        self.widget = widget
        self.setWindowTitle(window_title)
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 0, 6, 6)

        self.title_widget = QLabel()
        self.title_widget.setObjectName("fieldTitle")
        self.title_widget.setWordWrap(True)
        self.title_widget.setStyleSheet(
            "font-weight:bold;" "font-size:14px;" "padding-top: 6px"
        )
        self.title_widget.setText(title)
        self.title_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.button_container = QWidget()
        self.button_layout = QHBoxLayout(self.button_container)
        self.button_layout.setContentsMargins(6, 6, 6, 6)
        self.button_layout.addStretch(1)

        # self.cancel_button = QPushButton()
        # self.cancel_button.setText('Cancel')

        if not (save_callback or has_save):
            self.done_button = QPushButton()
            self.done_button.setText("Done")
            self.done_button.setAutoDefault(True)
            self.done_button.clicked.connect(self.hide)
            if done_callback:
                self.done_button.clicked.connect(done_callback)
            self.button_layout.addWidget(self.done_button)

        if save_callback or has_save:
            self.cancel_button = QPushButton()
            self.cancel_button.setText("Cancel")
            self.cancel_button.clicked.connect(self.hide)
            self.cancel_button.clicked.connect(widget.reset)
            # self.cancel_button.clicked.connect(cancel_callback)
            self.button_layout.addWidget(self.cancel_button)

            self.save_button = QPushButton()
            self.save_button.setText("Save")
            self.save_button.setAutoDefault(True)
            self.save_button.clicked.connect(self.hide)
            self.save_button.clicked.connect(self.saved.emit)

            if done_callback:
                self.save_button.clicked.connect(done_callback)
            if save_callback:
                self.save_button.clicked.connect(
                    lambda: save_callback(widget.get_content())
                )
            self.button_layout.addWidget(self.save_button)

            # trigger save button actions when pressing enter in the widget
            self.widget.add_callback(lambda: self.save_button.click())

        widget.done.connect(lambda: save_callback(widget.get_content()))

        self.root_layout.addWidget(self.title_widget)
        self.root_layout.addWidget(widget)
        self.root_layout.setStretch(1, 2)
        self.root_layout.addWidget(self.button_container)


class PanelWidget(QWidget):
    """
    Used for widgets that go in a modal panel, ex. for editing or searching.
    """

    done = Signal()

    def __init__(self):
        super().__init__()

    def get_content(self) -> str:
        pass

    def reset(self):
        pass

    def add_callback(self, callback: Callable, event: str = "returnPressed"):
        logging.warning(f"add_callback not implemented for {self.__class__.__name__}")



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

"""A pagination widget created for TagStudio."""
# I never want to see this code again.

from PySide6.QtCore import QObject, Signal, QSize
from PySide6.QtGui import QIntValidator
from PySide6.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QPushButton,
    QLabel,
    QLineEdit,
    QSizePolicy,
)
from src.qt.helpers.qbutton_wrapper import QPushButtonWrapper

# class NumberEdit(QLineEdit):
# 	def __init__(self, parent=None) -> None:
# 		super().__init__(parent)
# 		self.textChanged


class Pagination(QWidget, QObject):
    """Widget containing controls for navigating between pages of items."""

    index = Signal(int)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.page_count: int = 0
        self.current_page_index: int = 0
        self.buffer_page_count: int = 4
        self.button_size = QSize(32, 24)

        # ------------ UI EXAMPLE --------------
        # [<] [1]...[3][4] [5] [6][7]...[42] [>]
        #            ^^^^ <-- 2 Buffer Pages
        # Center Page Number is Editable Text
        # --------------------------------------

        # [----------- ROOT LAYOUT ------------]
        self.setHidden(True)
        self.root_layout = QHBoxLayout(self)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        self.root_layout.setContentsMargins(0, 6, 0, 0)
        self.root_layout.setSpacing(3)
        # self.setMinimumHeight(32)

        # [<] ----------------------------------
        self.prev_button = QPushButtonWrapper()
        self.prev_button.setText("<")
        self.prev_button.setMinimumSize(self.button_size)
        self.prev_button.setMaximumSize(self.button_size)

        # --- [1] ------------------------------
        self.start_button = QPushButtonWrapper()
        self.start_button.setMinimumSize(self.button_size)
        self.start_button.setMaximumSize(self.button_size)
        # self.start_button.setStyleSheet('background:cyan;')
        # self.start_button.setMaximumHeight(self.button_size.height())

        # ------ ... ---------------------------
        self.start_ellipses = QLabel()
        self.start_ellipses.setMinimumSize(self.button_size)
        self.start_ellipses.setMaximumSize(self.button_size)
        # self.start_ellipses.setMaximumHeight(self.button_size.height())
        self.start_ellipses.setText(". . .")

        # --------- [3][4] ---------------------
        self.start_buffer_container = QWidget()
        self.start_buffer_layout = QHBoxLayout(self.start_buffer_container)
        self.start_buffer_layout.setContentsMargins(0, 0, 0, 0)
        self.start_buffer_layout.setSpacing(3)
        # self.start_buffer_container.setStyleSheet('background:blue;')

        # ---------------- [5] -----------------
        self.current_page_field = QLineEdit()
        self.current_page_field.setMinimumSize(self.button_size)
        self.current_page_field.setMaximumSize(self.button_size)
        self.validator = Validator(1, self.page_count)
        self.current_page_field.setValidator(self.validator)
        self.current_page_field.returnPressed.connect(
            lambda: self._goto_page(int(self.current_page_field.text()) - 1)
        )
        # self.current_page_field.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        # self.current_page_field.setMaximumHeight(self.button_size.height())
        # self.current_page_field.setMaximumWidth(self.button_size.width())

        # -------------------- [6][7] ----------
        self.end_buffer_container = QWidget()
        self.end_buffer_layout = QHBoxLayout(self.end_buffer_container)
        self.end_buffer_layout.setContentsMargins(0, 0, 0, 0)
        self.end_buffer_layout.setSpacing(3)
        # self.end_buffer_container.setStyleSheet('background:orange;')

        # -------------------------- ... -------
        self.end_ellipses = QLabel()
        self.end_ellipses.setMinimumSize(self.button_size)
        self.end_ellipses.setMaximumSize(self.button_size)
        # self.end_ellipses.setMaximumHeight(self.button_size.height())
        self.end_ellipses.setText(". . .")

        # ----------------------------- [42] ---
        self.end_button = QPushButtonWrapper()
        self.end_button.setMinimumSize(self.button_size)
        self.end_button.setMaximumSize(self.button_size)
        # self.end_button.setMaximumHeight(self.button_size.height())
        # self.end_button.setStyleSheet('background:red;')

        # ---------------------------------- [>]
        self.next_button = QPushButtonWrapper()
        self.next_button.setText(">")
        self.next_button.setMinimumSize(self.button_size)
        self.next_button.setMaximumSize(self.button_size)

        # Add Widgets to Root Layout
        self.root_layout.addStretch(1)
        self.root_layout.addWidget(self.prev_button)
        self.root_layout.addWidget(self.start_button)
        self.root_layout.addWidget(self.start_ellipses)
        self.root_layout.addWidget(self.start_buffer_container)
        self.root_layout.addWidget(self.current_page_field)
        self.root_layout.addWidget(self.end_buffer_container)
        self.root_layout.addWidget(self.end_ellipses)
        self.root_layout.addWidget(self.end_button)
        self.root_layout.addWidget(self.next_button)
        self.root_layout.addStretch(1)

        self._populate_buffer_buttons()
        # self.update_buttons(page_count=9, index=0)

    def update_buttons(self, page_count: int, index: int, emit: bool = True):
        # Guard
        if index < 0:
            raise ValueError("Negative index detected")

        # Screw it
        for i in range(0, 10):
            if self.start_buffer_layout.itemAt(i):
                self.start_buffer_layout.itemAt(i).widget().setHidden(True)
            if self.end_buffer_layout.itemAt(i):
                self.end_buffer_layout.itemAt(i).widget().setHidden(True)

        end_page = page_count - 1
        if page_count <= 1:
            # Hide everything if there are only one or less pages.
            # [-------------- HIDDEN --------------]
            self.setHidden(True)
        # elif page_count > 1 and page_count < 7:
        # 	# Only show Next/Prev, current index field, and both start and end
        # 	# buffers (the end may be odd).
        # 	# [<] [1][2][3][4][5][6] [>]
        # 	self.start_button.setHidden(True)
        # 	self.start_ellipses.setHidden(True)
        # 	self.end_ellipses.setHidden(True)
        # 	self.end_button.setHidden(True)
        # elif page_count > 1:
        # 	self.start_button.setHidden(False)
        # 	self.start_ellipses.setHidden(False)
        # 	self.end_ellipses.setHidden(False)
        # 	self.end_button.setHidden(False)

        # 	self.start_button.setText('1')
        # 	self.assign_click(self.start_button, 0)
        # 	self.end_button.setText(str(page_count))
        # 	self.assign_click(self.end_button, page_count-1)

        elif page_count > 1:
            # Enable/Disable Next+Prev Buttons
            if index == 0:
                self.prev_button.setDisabled(True)
                # self.start_buffer_layout.setContentsMargins(0,0,0,0)
            else:
                # self.start_buffer_layout.setContentsMargins(3,0,3,0)
                self._assign_click(self.prev_button, index - 1)
                self.prev_button.setDisabled(False)

            if index == end_page:
                self.next_button.setDisabled(True)
                # self.end_buffer_layout.setContentsMargins(0,0,0,0)
            else:
                # self.end_buffer_layout.setContentsMargins(3,0,3,0)
                self._assign_click(self.next_button, index + 1)
                self.next_button.setDisabled(False)

            # Set Ellipses Sizes
            # I do not know where these magic values were derived from, but
            # this is better than the chain elif's that were here before
            if 8 <= page_count <= 11:
                end_scale = max(1, page_count - index - 6)
                srt_scale = max(1, index - 5)
            elif page_count > 11:
                end_scale = max(1, 7 - index)
                srt_scale = max(1, (7 - (end_page - index)))

            if page_count >= 8:
                end_size = self.button_size.width() * end_scale + (3 * (end_scale - 1))
                srt_size = self.button_size.width() * srt_scale + (3 * (srt_scale - 1))
                self.end_ellipses.setMinimumWidth(end_size)
                self.end_ellipses.setMaximumWidth(end_size)
                self.start_ellipses.setMinimumWidth(srt_size)
                self.start_ellipses.setMaximumWidth(srt_size)

            # Enable/Disable Ellipses
            # if index <= max(self.buffer_page_count, 5)+1:
            if index <= self.buffer_page_count + 1:
                self.start_ellipses.setHidden(True)
                # self.start_button.setHidden(True)
            else:
                self.start_ellipses.setHidden(False)
                # self.start_button.setHidden(False)
                # self.start_button.setText('1')
                self._assign_click(self.start_button, 0)
            # if index >=(page_count-max(self.buffer_page_count, 5)-2):
            if index >= (page_count - self.buffer_page_count - 2):
                self.end_ellipses.setHidden(True)
                # self.end_button.setHidden(True)
            else:
                self.end_ellipses.setHidden(False)
                # self.end_button.setHidden(False)
                # self.end_button.setText(str(page_count))
                # self.assign_click(self.end_button, page_count-1)

            # Hide/Unhide Start+End Buttons
            if index != 0:
                self.start_button.setText("1")
                self._assign_click(self.start_button, 0)
                self.start_button.setHidden(False)
                # self.start_buffer_layout.setContentsMargins(3,0,0,0)
            else:
                self.start_button.setHidden(True)
                # self.start_buffer_layout.setContentsMargins(0,0,0,0)
            if index != page_count - 1:
                self.end_button.setText(str(page_count))
                self._assign_click(self.end_button, page_count - 1)
                self.end_button.setHidden(False)
                # self.end_buffer_layout.setContentsMargins(0,0,3,0)
            else:
                self.end_button.setHidden(True)
                # self.end_buffer_layout.setContentsMargins(0,0,0,0)

            if index == 0 or index == 1:
                self.start_buffer_container.setHidden(True)
            else:
                self.start_buffer_container.setHidden(False)

            if index == page_count - 1 or index == page_count - 2:
                self.end_buffer_container.setHidden(True)
            else:
                self.end_buffer_container.setHidden(False)

            # for i in range(0, self.buffer_page_count):
            # 	self.start_buffer_layout.itemAt(i).widget().setHidden(True)

            # Current Field and Buffer Pages
            sbc = 0
            # for i in range(0, max(self.buffer_page_count*2, 11)):
            for i in range(0, page_count):
                # for j in range(0, self.buffer_page_count+1):
                # 	self.start_buffer_layout.itemAt(j).widget().setHidden(True)
                # if i == 1:
                # 	self.start_buffer_layout.itemAt(i).widget().setHidden(True)
                # elif i == page_count-2:
                # 	self.end_buffer_layout.itemAt(i).widget().setHidden(True)

                # Set Field
                if i == index:
                    # print(f'Current Index: {i}')
                    if self.start_buffer_layout.itemAt(i):
                        self.start_buffer_layout.itemAt(i).widget().setHidden(True)
                    if self.end_buffer_layout.itemAt(i):
                        self.end_buffer_layout.itemAt(i).widget().setHidden(True)
                    sbc += 1
                    self.current_page_field.setText((str(i + 1)))
                # elif index == page_count-1:
                # 	self.start_button.setText(str(page_count))

                start_offset = max(0, (index - 4) - 4)
                end_offset = min(page_count - 1, (index + 4) - 4)
                if i < index:
                    # if i != 0 and ((i-self.buffer_page_count) >= 0 or i <= self.buffer_page_count):
                    if (i != 0) and i >= index - 4:
                        # print(f'     Start i: {i}')
                        # print(f'Start Offset: {start_offset}')
                        # print(f' Requested i: {i-start_offset}')
                        # print(f'Setting Text "{str(i+1)}" for Local Start i:{i-start_offset}, Global i:{i}')
                        self.start_buffer_layout.itemAt(
                            i - start_offset
                        ).widget().setHidden(False)
                        self.start_buffer_layout.itemAt(
                            i - start_offset
                        ).widget().setText(str(i + 1))  # type: ignore
                        self._assign_click(
                            self.start_buffer_layout.itemAt(i - start_offset).widget(),  # type: ignore
                            i,
                        )
                        sbc += 1
                    else:
                        if self.start_buffer_layout.itemAt(i):
                            # print(f'Removing S-Start {i}')
                            self.start_buffer_layout.itemAt(i).widget().setHidden(True)
                        if self.end_buffer_layout.itemAt(i):
                            # print(f'Removing S-End {i}')
                            self.end_buffer_layout.itemAt(i).widget().setHidden(True)
                elif i > index:
                    # if i != page_count-1:
                    if i != page_count - 1 and i <= index + 4:
                        # print(f'End Buffer: {i}')
                        # print(f'      End i: {i}')
                        # print(f' End Offset: {end_offset}')
                        # print(f'Requested i: {i-end_offset}')
                        # print(f'Requested i: {end_offset-sbc-i}')
                        # if self.start_buffer_layout.itemAt(i):
                        # 	self.start_buffer_layout.itemAt(i).widget().setHidden(True)
                        # print(f'Setting Text "{str(i+1)}" for Local End i:{i-end_offset}, Global i:{i}')
                        self.end_buffer_layout.itemAt(
                            i - end_offset
                        ).widget().setHidden(False)
                        self.end_buffer_layout.itemAt(i - end_offset).widget().setText(  # type: ignore
                            str(i + 1)
                        )
                        self._assign_click(
                            self.end_buffer_layout.itemAt(i - end_offset).widget(),  # type: ignore
                            i,
                        )
                    else:
                        # if self.start_buffer_layout.itemAt(i-1):
                        # 	print(f'Removing E-Start {i-1}')
                        # 	self.start_buffer_layout.itemAt(i-1).widget().setHidden(True)
                        # if self.start_buffer_layout.itemAt(i-start_offset):
                        # 	print(f'Removing E-Start Offset {i-end_offset}')
                        # 	self.start_buffer_layout.itemAt(i-end_offset).widget().setHidden(True)

                        if self.end_buffer_layout.itemAt(i):
                            # print(f'Removing E-End {i}')
                            self.end_buffer_layout.itemAt(i).widget().setHidden(True)
                        for j in range(0, self.buffer_page_count):
                            if self.end_buffer_layout.itemAt(i - end_offset + j):
                                # print(f'Removing E-End-Offset {i-end_offset+j}')
                                self.end_buffer_layout.itemAt(
                                    i - end_offset + j
                                ).widget().setHidden(True)

                    # if self.end_buffer_layout.itemAt(i+1):
                    # 	print(f'Removing T-End {i+1}')
                    # 	self.end_buffer_layout.itemAt(i+1).widget().setHidden(True)

                    if self.start_buffer_layout.itemAt(i - 1):
                        # print(f'Removing T-Start {i-1}')
                        self.start_buffer_layout.itemAt(i - 1).widget().setHidden(True)

                # if index == 0 or index == 1:
                # 	print(f'Removing Start i: {i}')
                # 	if self.start_buffer_layout.itemAt(i):
                # 		self.start_buffer_layout.itemAt(i).widget().setHidden(True)

                # elif index == page_count-1 or index == page_count-2 or index == page_count-3 or index == page_count-4:
                # 	print(f' Removing End i: {i}')
                # 	if self.end_buffer_layout.itemAt(i):
                # 		self.end_buffer_layout.itemAt(i).widget().setHidden(True)

                # else:
                # 	print(f'Truncate: {i}')
                # 	if self.start_buffer_layout.itemAt(i):
                # 		self.start_buffer_layout.itemAt(i).widget().setHidden(True)
                # 	if self.end_buffer_layout.itemAt(i):
                # 		self.end_buffer_layout.itemAt(i).widget().setHidden(True)

                # if i < self.buffer_page_count:
                # 	print(f'start {i}')
                # 	if i == 0:
                # 		self.start_buffer_layout.itemAt(i).widget().setHidden(True)
                # 		self.current_page_field.setText((str(i+1)))
                # 	else:
                # 		self.start_buffer_layout.itemAt(i).widget().setHidden(False)
                # 		self.start_buffer_layout.itemAt(i).widget().setText(str(i+1))
                # elif i >= self.buffer_page_count and i < count:
                # 	print(f'end {i}')
                # 	self.end_buffer_layout.itemAt(i-self.buffer_page_count).widget().setHidden(False)
                # 	self.end_buffer_layout.itemAt(i-self.buffer_page_count).widget().setText(str(i+1))
                # else:
                # 	self.end_buffer_layout.itemAt(i-self.buffer_page_count).widget().setHidden(True)

            self.setHidden(False)
        # elif page_count >= 7:
        # 	# Show everything, except truncate the buffers as needed.
        # 	# [<] [1]...[3] [4] [5]...[7] [>]
        # 	self.start_button.setHidden(False)
        # 	self.start_ellipses.setHidden(False)
        # 	self.end_ellipses.setHidden(False)
        # 	self.end_button.setHidden(False)

        # 	if index == 0:
        # 		self.prev_button.setDisabled(True)
        # 		self.start_buffer_layout.setContentsMargins(0,0,3,0)
        # 	else:
        # 		self.start_buffer_layout.setContentsMargins(3,0,3,0)
        # 		self.assign_click(self.prev_button, index-1)
        # 		self.prev_button.setDisabled(False)

        # 	if index == page_count-1:
        # 		self.next_button.setDisabled(True)
        # 		self.end_buffer_layout.setContentsMargins(3,0,0,0)
        # 	else:
        # 		self.end_buffer_layout.setContentsMargins(3,0,3,0)
        # 		self.assign_click(self.next_button, index+1)
        # 		self.next_button.setDisabled(False)

        # 	self.start_button.setText('1')
        # 	self.assign_click(self.start_button, 0)
        # 	self.end_button.setText(str(page_count))
        # 	self.assign_click(self.end_button, page_count-1)

        # 	self.setHidden(False)

        self.validator.setTop(page_count)
        # if self.current_page_index != index:
        if emit:
            print(f"[PAGINATION] Emitting {index}")
            self.index.emit(index)
        self.current_page_index = index
        self.page_count = page_count

    def _goto_page(self, index: int):
        # print(f'GOTO PAGE: {index}')
        self.update_buttons(self.page_count, index)

    def _assign_click(self, button: QPushButtonWrapper, index):
        if button.is_connected:
            button.clicked.disconnect()
        button.clicked.connect(lambda checked=False, i=index: self._goto_page(i))
        button.is_connected = True

    def _populate_buffer_buttons(self):
        for i in range(max(self.buffer_page_count * 2, 5)):
            button = QPushButtonWrapper()
            button.setMinimumSize(self.button_size)
            button.setMaximumSize(self.button_size)
            button.setHidden(True)
            # button.setMaximumHeight(self.button_size.height())
            self.start_buffer_layout.addWidget(button)

        for i in range(max(self.buffer_page_count * 2, 5)):
            button = QPushButtonWrapper()
            button.setMinimumSize(self.button_size)
            button.setMaximumSize(self.button_size)
            button.setHidden(True)
            # button.setMaximumHeight(self.button_size.height())
            self.end_buffer_layout.addWidget(button)


class Validator(QIntValidator):
    def __init__(self, bottom: int, top: int, parent=None) -> None:
        super().__init__(bottom, top, parent)

    def fixup(self, input: str) -> str:
        # print(input)
        input = input.strip("0")
        print(input)
        return super().fixup(str(self.top()) if input else "1")



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

from chardet.universaldetector import UniversalDetector
from pathlib import Path


def detect_char_encoding(filepath: Path) -> str | None:
    """
    Attempts to detect the character encoding of a text file.

    Args:
    filepath (Path): The path of the text file to analyze.

    Returns:
    str | None: The detected character encoding, if any.
    """

    detector = UniversalDetector()
    with open(filepath, "rb") as text_file:
        for line in text_file.readlines():
            detector.feed(line)
            if detector.done:
                break
        detector.close()
    return detector.result["encoding"]



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


def strip_web_protocol(string: str) -> str:
    """Strips a leading web protocol (ex. \"https://\") as well as \"www.\" from a string."""
    prefixes = ["https://", "http://", "www.", "www2."]
    for prefix in prefixes:
        string = string.removeprefix(prefix)
    return string



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

import logging
from pathlib import Path
from typing import Any

import ujson

logging.basicConfig(format="%(message)s", level=logging.INFO)


class ResourceManager:
    """A resource manager for retrieving resources."""

    _map: dict = {}
    _cache: dict[str, Any] = {}
    _initialized: bool = False

    def __init__(self) -> None:
        # Load JSON resource map
        if not ResourceManager._initialized:
            with open(
                Path(__file__).parent / "resources.json", mode="r", encoding="utf-8"
            ) as f:
                ResourceManager._map = ujson.load(f)
                logging.info(
                    f"[ResourceManager] {len(ResourceManager._map.items())} resources registered"
                )
            ResourceManager._initialized = True

    def get(self, id: str) -> Any:
        """Get a resource from the ResourceManager.
        This can include resources inside and outside of QResources, and will return
        theme-respecting variations of resources if available.

        Args:
            id (str): The name of the resource.

        Returns:
            Any: The resource if found, else None.
        """
        cached_res = ResourceManager._cache.get(id)
        if cached_res:
            return cached_res
        else:
            res: dict = ResourceManager._map.get(id)
            if res.get("mode") in ["r", "rb"]:
                with open(
                    (Path(__file__).parents[2] / "resources" / res.get("path")),
                    res.get("mode"),
                ) as f:
                    data = f.read()
                    if res.get("mode") == "rb":
                        data = bytes(data)
                    ResourceManager._cache[id] = data
                    return data
            elif res.get("mode") in ["qt"]:
                # TODO: Qt resource loading logic
                pass

    def __getattr__(self, __name: str) -> Any:
        attr = self.get(__name)
        if attr:
            return attr
        raise AttributeError(f"Attribute {id} not found")



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

import logging
import os
import traceback
from pathlib import Path

import cv2
from PIL import Image, ImageChops, UnidentifiedImageError
from PIL.Image import DecompressionBombError
from PySide6.QtCore import (
    QObject,
    QThread,
    Signal,
    QRunnable,
    Qt,
    QThreadPool,
    QSize,
    QEvent,
    QTimer,
    QSettings,
)

from src.core.library import Library
from src.core.constants import DOC_TYPES, VIDEO_TYPES, IMAGE_TYPES


ERROR = f"[ERROR]"
WARNING = f"[WARNING]"
INFO = f"[INFO]"


logging.basicConfig(format="%(message)s", level=logging.INFO)


class CollageIconRenderer(QObject):
    rendered = Signal(Image.Image)
    done = Signal()

    def __init__(self, library: Library):
        QObject.__init__(self)
        self.lib = library

    def render(
        self,
        entry_id,
        size: tuple[int, int],
        data_tint_mode,
        data_only_mode,
        keep_aspect,
    ):
        entry = self.lib.get_entry(entry_id)
        filepath = self.lib.library_dir / entry.path / entry.filename
        file_type = os.path.splitext(filepath)[1].lower()[1:]
        color: str = ""

        try:
            if data_tint_mode or data_only_mode:
                color = "#000000"  # Black (Default)

                if entry.fields:
                    has_any_tags: bool = False
                    has_content_tags: bool = False
                    has_meta_tags: bool = False
                    for field in entry.fields:
                        if self.lib.get_field_attr(field, "type") == "tag_box":
                            if self.lib.get_field_attr(field, "content"):
                                has_any_tags = True
                                if self.lib.get_field_attr(field, "id") == 7:
                                    has_content_tags = True
                                elif self.lib.get_field_attr(field, "id") == 8:
                                    has_meta_tags = True
                    if has_content_tags and has_meta_tags:
                        color = "#28bb48"  # Green
                    elif has_any_tags:
                        color = "#ffd63d"  # Yellow
                        # color = '#95e345' # Yellow-Green
                    else:
                        # color = '#fa9a2c' # Yellow-Orange
                        color = "#ed8022"  # Orange
                else:
                    color = "#e22c3c"  # Red

                if data_only_mode:
                    pic = Image.new("RGB", size, color)
                    # collage.paste(pic, (y*thumb_size, x*thumb_size))
                    self.rendered.emit(pic)
            if not data_only_mode:
                logging.info(
                    f"\r{INFO} Combining [ID:{entry_id}/{len(self.lib.entries)}]: {self.get_file_color(filepath.suffix.lower())}{entry.path}{os.sep}{entry.filename}\033[0m"
                )
                # sys.stdout.write(f'\r{INFO} Combining [{i+1}/{len(self.lib.entries)}]: {self.get_file_color(file_type)}{entry.path}{os.sep}{entry.filename}{RESET}')
                # sys.stdout.flush()
                if filepath.suffix.lower() in IMAGE_TYPES:
                    try:
                        with Image.open(
                            str(self.lib.library_dir / entry.path / entry.filename)
                        ) as pic:
                            if keep_aspect:
                                pic.thumbnail(size)
                            else:
                                pic = pic.resize(size)
                            if data_tint_mode and color:
                                pic = pic.convert(mode="RGB")
                                pic = ImageChops.hard_light(
                                    pic, Image.new("RGB", size, color)
                                )
                            # collage.paste(pic, (y*thumb_size, x*thumb_size))
                            self.rendered.emit(pic)
                    except DecompressionBombError as e:
                        logging.info(f"[ERROR] One of the images was too big ({e})")
                elif filepath.suffix.lower() in VIDEO_TYPES:
                    video = cv2.VideoCapture(str(filepath))
                    video.set(
                        cv2.CAP_PROP_POS_FRAMES,
                        (video.get(cv2.CAP_PROP_FRAME_COUNT) // 2),
                    )
                    success, frame = video.read()
                    if not success:
                        # Depending on the video format, compression, and frame
                        # count, seeking halfway does not work and the thumb
                        # must be pulled from the earliest available frame.
                        video.set(cv2.CAP_PROP_POS_FRAMES, 0)
                        success, frame = video.read()
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    with Image.fromarray(frame, mode="RGB") as pic:
                        if keep_aspect:
                            pic.thumbnail(size)
                        else:
                            pic = pic.resize(size)
                        if data_tint_mode and color:
                            pic = ImageChops.hard_light(
                                pic, Image.new("RGB", size, color)
                            )
                        # collage.paste(pic, (y*thumb_size, x*thumb_size))
                        self.rendered.emit(pic)
        except (UnidentifiedImageError, FileNotFoundError):
            logging.info(
                f"\n{ERROR} Couldn't read {entry.path}{os.sep}{entry.filename}"
            )
            with Image.open(
                str(
                    Path(__file__).parents[2]
                    / "resources/qt/images/thumb_broken_512.png"
                )
            ) as pic:
                pic.thumbnail(size)
                if data_tint_mode and color:
                    pic = pic.convert(mode="RGB")
                    pic = ImageChops.hard_light(pic, Image.new("RGB", size, color))
                # collage.paste(pic, (y*thumb_size, x*thumb_size))
                self.rendered.emit(pic)
        except KeyboardInterrupt:
            # self.quit(save=False, backup=True)
            run = False
            # clear()
            logging.info("\n")
            logging.info(f"{INFO} Collage operation cancelled.")
            clear_scr = False
        except:
            logging.info(f"{ERROR} {entry.path}{os.sep}{entry.filename}")
            traceback.print_exc()
            logging.info("Continuing...")

        self.done.emit()
        # logging.info('Done!')

    def get_file_color(self, ext: str):
        if ext.lower().replace(".", "", 1) == "gif":
            return "\033[93m"
        if ext.lower().replace(".", "", 1) in IMAGE_TYPES:
            return "\033[37m"
        elif ext.lower().replace(".", "", 1) in VIDEO_TYPES:
            return "\033[96m"
        elif ext.lower().replace(".", "", 1) in DOC_TYPES:
            return "\033[92m"
        else:
            return "\033[97m"



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from PySide6.QtCore import Signal, QObject
from typing import Callable


class FunctionIterator(QObject):
    """Iterates over a yielding function and emits progress as the 'value' signal.\n\nThread-Safe Guarantee™"""

    value = Signal(object)

    def __init__(self, function: Callable):
        super().__init__()
        self.iterable = function

    def run(self):
        for i in self.iterable():
            self.value.emit(i)



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QLabel


class ClickableLabel(QLabel):
    """A clickable Label widget."""

    clicked = Signal()

    def __init__(self):
        super().__init__()

    def mousePressEvent(self, event):
        self.clicked.emit()



VERSION: str = "9.3.2"  # Major.Minor.Patch
VERSION_BRANCH: str = ""  # Usually "" or "Pre-Release"

# The folder & file names where TagStudio keeps its data relative to a library.
TS_FOLDER_NAME: str = ".TagStudio"
BACKUP_FOLDER_NAME: str = "backups"
COLLAGE_FOLDER_NAME: str = "collages"
LIBRARY_FILENAME: str = "ts_library.json"

# TODO: Turn this whitelist into a user-configurable blacklist.
IMAGE_TYPES: list[str] = [
    ".png",
    ".jpg",
    ".jpeg",
    ".jpg_large",
    ".jpeg_large",
    ".jfif",
    ".gif",
    ".tif",
    ".tiff",
    ".heic",
    ".heif",
    ".webp",
    ".bmp",
    ".svg",
    ".avif",
    ".apng",
    ".jp2",
    ".j2k",
    ".jpg2",
]
RAW_IMAGE_TYPES: list[str] = [
    ".raw",
    ".dng",
    ".rw2",
    ".nef",
    ".arw",
    ".crw",
    ".cr2",
    ".cr3",
]
VIDEO_TYPES: list[str] = [
    ".mp4",
    ".webm",
    ".mov",
    ".hevc",
    ".mkv",
    ".avi",
    ".wmv",
    ".flv",
    ".gifv",
    ".m4p",
    ".m4v",
    ".3gp",
]
AUDIO_TYPES: list[str] = [
    ".mp3",
    ".mp4",
    ".mpeg4",
    ".m4a",
    ".aac",
    ".wav",
    ".flac",
    ".alac",
    ".wma",
    ".ogg",
    ".aiff",
]
DOC_TYPES: list[str] = [
    ".txt",
    ".rtf",
    ".md",
    ".doc",
    ".docx",
    ".pdf",
    ".tex",
    ".odt",
    ".pages",
]
PLAINTEXT_TYPES: list[str] = [
    ".txt",
    ".md",
    ".css",
    ".html",
    ".xml",
    ".json",
    ".js",
    ".ts",
    ".ini",
    ".htm",
    ".csv",
    ".php",
    ".sh",
    ".bat",
]
SPREADSHEET_TYPES: list[str] = [".csv", ".xls", ".xlsx", ".numbers", ".ods"]
PRESENTATION_TYPES: list[str] = [".ppt", ".pptx", ".key", ".odp"]
ARCHIVE_TYPES: list[str] = [
    ".zip",
    ".rar",
    ".tar",
    ".tar",
    ".gz",
    ".tgz",
    ".7z",
    ".s7z",
]
PROGRAM_TYPES: list[str] = [".exe", ".app"]
SHORTCUT_TYPES: list[str] = [".lnk", ".desktop", ".url"]

ALL_FILE_TYPES: list[str] = (
    IMAGE_TYPES
    + VIDEO_TYPES
    + AUDIO_TYPES
    + DOC_TYPES
    + SPREADSHEET_TYPES
    + PRESENTATION_TYPES
    + ARCHIVE_TYPES
    + PROGRAM_TYPES
    + SHORTCUT_TYPES
)

BOX_FIELDS = ["tag_box", "text_box"]
TEXT_FIELDS = ["text_line", "text_box"]
DATE_FIELDS = ["datetime"]

TAG_COLORS = [
    "",
    "black",
    "dark gray",
    "gray",
    "light gray",
    "white",
    "light pink",
    "pink",
    "red",
    "red orange",
    "orange",
    "yellow orange",
    "yellow",
    "lime",
    "light green",
    "mint",
    "green",
    "teal",
    "cyan",
    "light blue",
    "blue",
    "blue violet",
    "violet",
    "purple",
    "lavender",
    "berry",
    "magenta",
    "salmon",
    "auburn",
    "dark brown",
    "brown",
    "light brown",
    "blonde",
    "peach",
    "warm gray",
    "cool gray",
    "olive",
]

TAG_FAVORITE = 1
TAG_ARCHIVED = 0



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio
import contextlib
import logging
import os
import time
import typing
from pathlib import Path
from typing import Optional

from PIL import Image, ImageQt
from PySide6.QtCore import Qt, QSize, QEvent
from PySide6.QtGui import QPixmap, QEnterEvent, QAction
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QBoxLayout,
    QCheckBox,
)

from src.core.enums import FieldID
from src.core.library import ItemType, Library, Entry
from src.core.constants import (
    AUDIO_TYPES,
    VIDEO_TYPES,
    IMAGE_TYPES,
    TAG_FAVORITE,
    TAG_ARCHIVED,
)
from src.qt.flowlayout import FlowWidget
from src.qt.helpers.file_opener import FileOpenerHelper
from src.qt.widgets.thumb_renderer import ThumbRenderer
from src.qt.widgets.thumb_button import ThumbButton

if typing.TYPE_CHECKING:
    from src.qt.widgets.preview_panel import PreviewPanel

ERROR = f"[ERROR]"
WARNING = f"[WARNING]"
INFO = f"[INFO]"


logging.basicConfig(format="%(message)s", level=logging.INFO)


class ItemThumb(FlowWidget):
    """
    The thumbnail widget for a library item (Entry, Collation, Tag Group, etc.).
    """

    update_cutoff: float = time.time()

    collation_icon_128: Image.Image = Image.open(
        str(Path(__file__).parents[3] / "resources/qt/images/collation_icon_128.png")
    )
    collation_icon_128.load()

    tag_group_icon_128: Image.Image = Image.open(
        str(Path(__file__).parents[3] / "resources/qt/images/tag_group_icon_128.png")
    )
    tag_group_icon_128.load()

    small_text_style = (
        f"background-color:rgba(0, 0, 0, 192);"
        f"font-family:Oxanium;"
        f"font-weight:bold;"
        f"font-size:12px;"
        f"border-radius:3px;"
        f"padding-top: 4px;"
        f"padding-right: 1px;"
        f"padding-bottom: 1px;"
        f"padding-left: 1px;"
    )

    med_text_style = (
        f"background-color:rgba(0, 0, 0, 192);"
        f"font-family:Oxanium;"
        f"font-weight:bold;"
        f"font-size:18px;"
        f"border-radius:3px;"
        f"padding-top: 4px;"
        f"padding-right: 1px;"
        f"padding-bottom: 1px;"
        f"padding-left: 1px;"
    )

    def __init__(
        self,
        mode: Optional[ItemType],
        library: Library,
        panel: "PreviewPanel",
        thumb_size: tuple[int, int],
    ):
        """Modes: entry, collation, tag_group"""
        super().__init__()
        self.lib = library
        self.panel = panel
        self.mode = mode
        self.item_id: int = -1
        self.isFavorite: bool = False
        self.isArchived: bool = False
        self.thumb_size: tuple[int, int] = thumb_size
        self.setMinimumSize(*thumb_size)
        self.setMaximumSize(*thumb_size)
        check_size = 24
        # self.setStyleSheet('background-color:red;')

        # +----------+
        # |   ARC FAV| Top Right: Favorite & Archived Badges
        # |          |
        # |          |
        # |EXT      #| Lower Left: File Type, Tag Group Icon, or Collation Icon
        # +----------+ Lower Right: Collation Count, Video Length, or Word Count

        # Thumbnail ============================================================

        # +----------+
        # |*--------*|
        # ||        ||
        # ||        ||
        # |*--------*|
        # +----------+
        self.base_layout = QVBoxLayout(self)
        self.base_layout.setObjectName("baseLayout")
        # self.base_layout.setRowStretch(1, 2)
        self.base_layout.setContentsMargins(0, 0, 0, 0)

        # +----------+
        # |[~~~~~~~~]|
        # |          |
        # |          |
        # |          |
        # +----------+
        self.top_layout = QHBoxLayout()
        self.top_layout.setObjectName("topLayout")
        # self.top_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        # self.top_layout.setColumnStretch(1, 2)
        self.top_layout.setContentsMargins(6, 6, 6, 6)
        self.top_container = QWidget()
        self.top_container.setLayout(self.top_layout)
        self.base_layout.addWidget(self.top_container)

        # +----------+
        # |[~~~~~~~~]|
        # |     ^    |
        # |     |    |
        # |     v    |
        # +----------+
        self.base_layout.addStretch(2)

        # +----------+
        # |[~~~~~~~~]|
        # |     ^    |
        # |     v    |
        # |[~~~~~~~~]|
        # +----------+
        self.bottom_layout = QHBoxLayout()
        self.bottom_layout.setObjectName("bottomLayout")
        # self.bottom_container.setAlignment(Qt.AlignmentFlag.AlignBottom)
        # self.bottom_layout.setColumnStretch(1, 2)
        self.bottom_layout.setContentsMargins(6, 6, 6, 6)
        self.bottom_container = QWidget()
        self.bottom_container.setLayout(self.bottom_layout)
        self.base_layout.addWidget(self.bottom_container)

        # self.root_layout = QGridLayout(self)
        # self.root_layout.setObjectName('rootLayout')
        # self.root_layout.setColumnStretch(1, 2)
        # self.root_layout.setRowStretch(1, 2)
        # self.root_layout.setContentsMargins(6,6,6,6)
        # # root_layout.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        self.thumb_button = ThumbButton(self, thumb_size)
        self.renderer = ThumbRenderer()
        self.renderer.updated.connect(
            lambda ts, i, s, ext: (
                self.update_thumb(ts, image=i),
                self.update_size(ts, size=s),
                self.set_extension(ext),  # type: ignore
            )
        )
        self.thumb_button.setFlat(True)

        # self.bg_button.setStyleSheet('background-color:blue;')
        # self.bg_button.setLayout(self.root_layout)
        self.thumb_button.setLayout(self.base_layout)
        # self.bg_button.setMinimumSize(*thumb_size)
        # self.bg_button.setMaximumSize(*thumb_size)

        self.thumb_button.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)
        self.opener = FileOpenerHelper("")
        open_file_action = QAction("Open file", self)
        open_file_action.triggered.connect(self.opener.open_file)
        open_explorer_action = QAction("Open file in explorer", self)
        open_explorer_action.triggered.connect(self.opener.open_explorer)
        self.thumb_button.addAction(open_file_action)
        self.thumb_button.addAction(open_explorer_action)

        # Static Badges ========================================================

        # Item Type Badge ------------------------------------------------------
        # Used for showing the Tag Group / Collation icons.
        # Mutually exclusive with the File Extension Badge.
        self.item_type_badge = QLabel()
        self.item_type_badge.setObjectName("itemBadge")
        self.item_type_badge.setPixmap(
            QPixmap.fromImage(
                ImageQt.ImageQt(
                    ItemThumb.collation_icon_128.resize(
                        (check_size, check_size), Image.Resampling.BILINEAR
                    )
                )
            )
        )
        self.item_type_badge.setMinimumSize(check_size, check_size)
        self.item_type_badge.setMaximumSize(check_size, check_size)
        # self.root_layout.addWidget(self.item_type_badge, 2, 0)
        self.bottom_layout.addWidget(self.item_type_badge)

        # File Extension Badge -------------------------------------------------
        # Mutually exclusive with the File Extension Badge.
        self.ext_badge = QLabel()
        self.ext_badge.setObjectName("extBadge")
        # self.ext_badge.setText('MP4')
        # self.ext_badge.setAlignment(Qt.AlignmentFlag.AlignVCenter)
        self.ext_badge.setStyleSheet(ItemThumb.small_text_style)
        # self.type_badge.setAlignment(Qt.AlignmentFlag.AlignRight)
        # self.root_layout.addWidget(self.ext_badge, 2, 0)
        self.bottom_layout.addWidget(self.ext_badge)
        # self.type_badge.setHidden(True)
        # bl_layout.addWidget(self.type_badge)

        self.bottom_layout.addStretch(2)

        # Count Badge ----------------------------------------------------------
        # Used for Tag Group + Collation counts, video length, word count, etc.
        self.count_badge = QLabel()
        self.count_badge.setObjectName("countBadge")
        # self.count_badge.setMaximumHeight(17)
        self.count_badge.setText("-:--")
        # self.count_badge.setAlignment(Qt.AlignmentFlag.AlignVCenter)
        self.count_badge.setStyleSheet(ItemThumb.small_text_style)
        # self.count_badge.setAlignment(Qt.AlignmentFlag.AlignBottom)
        # self.root_layout.addWidget(self.count_badge, 2, 2)
        self.bottom_layout.addWidget(
            self.count_badge, alignment=Qt.AlignmentFlag.AlignBottom
        )

        self.top_layout.addStretch(2)

        # Intractable Badges ===================================================
        self.cb_container = QWidget()
        # check_badges.setStyleSheet('background-color:cyan;')
        self.cb_layout = QHBoxLayout()
        self.cb_layout.setDirection(QBoxLayout.Direction.RightToLeft)
        self.cb_layout.setContentsMargins(0, 0, 0, 0)
        self.cb_layout.setSpacing(6)
        self.cb_container.setLayout(self.cb_layout)
        # self.cb_container.setHidden(True)
        # self.root_layout.addWidget(self.check_badges, 0, 2)
        self.top_layout.addWidget(self.cb_container)

        # Favorite Badge -------------------------------------------------------
        self.favorite_badge = QCheckBox()
        self.favorite_badge.setObjectName("favBadge")
        self.favorite_badge.setToolTip("Favorite")
        self.favorite_badge.setStyleSheet(
            f"QCheckBox::indicator{{width: {check_size}px;height: {check_size}px;}}"
            f"QCheckBox::indicator::unchecked{{image: url(:/images/star_icon_empty_128.png)}}"
            f"QCheckBox::indicator::checked{{image: url(:/images/star_icon_filled_128.png)}}"
            #  f'QCheckBox{{background-color:yellow;}}'
        )
        self.favorite_badge.setMinimumSize(check_size, check_size)
        self.favorite_badge.setMaximumSize(check_size, check_size)
        self.favorite_badge.stateChanged.connect(
            lambda x=self.favorite_badge.isChecked(): self.on_favorite_check(bool(x))
        )

        # self.fav_badge.setContentsMargins(0,0,0,0)
        # tr_layout.addWidget(self.fav_badge)
        # root_layout.addWidget(self.fav_badge, 0, 2)
        self.cb_layout.addWidget(self.favorite_badge)
        self.favorite_badge.setHidden(True)

        # Archive Badge --------------------------------------------------------
        self.archived_badge = QCheckBox()
        self.archived_badge.setObjectName("archiveBadge")
        self.archived_badge.setToolTip("Archive")
        self.archived_badge.setStyleSheet(
            f"QCheckBox::indicator{{width: {check_size}px;height: {check_size}px;}}"
            f"QCheckBox::indicator::unchecked{{image: url(:/images/box_icon_empty_128.png)}}"
            f"QCheckBox::indicator::checked{{image: url(:/images/box_icon_filled_128.png)}}"
            #  f'QCheckBox{{background-color:red;}}'
        )
        self.archived_badge.setMinimumSize(check_size, check_size)
        self.archived_badge.setMaximumSize(check_size, check_size)
        # self.archived_badge.clicked.connect(lambda x: self.assign_archived(x))
        self.archived_badge.stateChanged.connect(
            lambda x=self.archived_badge.isChecked(): self.on_archived_check(bool(x))
        )

        # tr_layout.addWidget(self.archive_badge)
        self.cb_layout.addWidget(self.archived_badge)
        self.archived_badge.setHidden(True)
        # root_layout.addWidget(self.archive_badge, 0, 2)
        # self.dumpObjectTree()

        self.set_mode(mode)

    def set_mode(self, mode: Optional[ItemType]) -> None:
        if mode is None:
            self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, True)
            self.unsetCursor()
            self.thumb_button.setHidden(True)
            # self.check_badges.setHidden(True)
            # self.ext_badge.setHidden(True)
            # self.item_type_badge.setHidden(True)
            pass
        elif mode == ItemType.ENTRY and self.mode != ItemType.ENTRY:
            self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, False)
            self.setCursor(Qt.CursorShape.PointingHandCursor)
            self.thumb_button.setHidden(False)
            self.cb_container.setHidden(False)
            # Count Badge depends on file extension (video length, word count)
            self.item_type_badge.setHidden(True)
            self.count_badge.setStyleSheet(ItemThumb.small_text_style)
            self.count_badge.setHidden(True)
            self.ext_badge.setHidden(True)
        elif mode == ItemType.COLLATION and self.mode != ItemType.COLLATION:
            self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, False)
            self.setCursor(Qt.CursorShape.PointingHandCursor)
            self.thumb_button.setHidden(False)
            self.cb_container.setHidden(True)
            self.ext_badge.setHidden(True)
            self.count_badge.setStyleSheet(ItemThumb.med_text_style)
            self.count_badge.setHidden(False)
            self.item_type_badge.setHidden(False)
        elif mode == ItemType.TAG_GROUP and self.mode != ItemType.TAG_GROUP:
            self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, False)
            self.setCursor(Qt.CursorShape.PointingHandCursor)
            self.thumb_button.setHidden(False)
            # self.cb_container.setHidden(True)
            self.ext_badge.setHidden(True)
            self.count_badge.setHidden(False)
            self.item_type_badge.setHidden(False)
        self.mode = mode
        # logging.info(f'Set Mode To: {self.mode}')

    # def update_(self, thumb: QPixmap, size:QSize, ext:str, badges:list[QPixmap]) -> None:
    # 	"""Updates the ItemThumb's visuals."""
    # 	if thumb:
    # 		pass

    def set_extension(self, ext: str) -> None:
        if ext and ext.startswith(".") is False:
            ext = "." + ext
        if ext and ext not in IMAGE_TYPES or ext in [".gif", ".apng"]:
            self.ext_badge.setHidden(False)
            self.ext_badge.setText(ext.upper()[1:])
            if ext in VIDEO_TYPES + AUDIO_TYPES:
                self.count_badge.setHidden(False)
        else:
            if self.mode == ItemType.ENTRY:
                self.ext_badge.setHidden(True)
                self.count_badge.setHidden(True)

    def set_count(self, count: str) -> None:
        if count:
            self.count_badge.setHidden(False)
            self.count_badge.setText(count)
        else:
            if self.mode == ItemType.ENTRY:
                self.ext_badge.setHidden(True)
                self.count_badge.setHidden(True)

    def update_thumb(self, timestamp: float, image: QPixmap = None):
        """Updates attributes of a thumbnail element."""
        # logging.info(f'[GUI] Updating Thumbnail for element {id(element)}: {id(image) if image else None}')
        if timestamp > ItemThumb.update_cutoff:
            self.thumb_button.setIcon(image if image else QPixmap())
            # element.repaint()

    def update_size(self, timestamp: float, size: QSize):
        """Updates attributes of a thumbnail element."""
        # logging.info(f'[GUI] Updating size for element {id(element)}:  {size.__str__()}')
        if timestamp > ItemThumb.update_cutoff:
            if self.thumb_button.iconSize != size:
                self.thumb_button.setIconSize(size)
                self.thumb_button.setMinimumSize(size)
                self.thumb_button.setMaximumSize(size)

    def update_clickable(self, clickable: typing.Callable):
        """Updates attributes of a thumbnail element."""
        # logging.info(f'[GUI] Updating Click Event for element {id(element)}: {id(clickable) if clickable else None}')
        if self.thumb_button.is_connected:
            self.thumb_button.clicked.disconnect()
        if clickable:
            self.thumb_button.clicked.connect(clickable)
            self.thumb_button.is_connected = True

    def update_badges(self):
        if self.mode == ItemType.ENTRY:
            # logging.info(f'[UPDATE BADGES] ENTRY: {self.lib.get_entry(self.item_id)}')
            # logging.info(f'[UPDATE BADGES] ARCH: {self.lib.get_entry(self.item_id).has_tag(self.lib, 0)}, FAV: {self.lib.get_entry(self.item_id).has_tag(self.lib, 1)}')
            self.assign_archived(
                self.lib.get_entry(self.item_id).has_tag(self.lib, TAG_ARCHIVED)
            )
            self.assign_favorite(
                self.lib.get_entry(self.item_id).has_tag(self.lib, TAG_FAVORITE)
            )

    def set_item_id(self, id: int):
        """
        also sets the filepath for the file opener
        """
        self.item_id = id
        if id == -1:
            return
        entry = self.lib.get_entry(self.item_id)
        filepath = self.lib.library_dir / entry.path / entry.filename
        self.opener.set_filepath(filepath)

    def assign_favorite(self, value: bool):
        # Switching mode to None to bypass mode-specific operations when the
        # checkbox's state changes.
        mode = self.mode
        self.mode = None
        self.isFavorite = value
        self.favorite_badge.setChecked(value)
        if not self.thumb_button.underMouse():
            self.favorite_badge.setHidden(not self.isFavorite)
        self.mode = mode

    def assign_archived(self, value: bool):
        # Switching mode to None to bypass mode-specific operations when the
        # checkbox's state changes.
        mode = self.mode
        self.mode = None
        self.isArchived = value
        self.archived_badge.setChecked(value)
        if not self.thumb_button.underMouse():
            self.archived_badge.setHidden(not self.isArchived)
        self.mode = mode

    def show_check_badges(self, show: bool):
        if self.mode != ItemType.TAG_GROUP:
            self.favorite_badge.setHidden(
                True if (not show and not self.isFavorite) else False
            )
            self.archived_badge.setHidden(
                True if (not show and not self.isArchived) else False
            )

    def enterEvent(self, event: QEnterEvent) -> None:
        self.show_check_badges(True)
        return super().enterEvent(event)

    def leaveEvent(self, event: QEvent) -> None:
        self.show_check_badges(False)
        return super().leaveEvent(event)

    def on_archived_check(self, toggle_value: bool):
        if self.mode == ItemType.ENTRY:
            self.isArchived = toggle_value
            self.toggle_item_tag(toggle_value, TAG_ARCHIVED)

    def on_favorite_check(self, toggle_value: bool):
        if self.mode == ItemType.ENTRY:
            self.isFavorite = toggle_value
            self.toggle_item_tag(toggle_value, TAG_FAVORITE)

    def toggle_item_tag(self, toggle_value: bool, tag_id: int):
        def toggle_tag(entry: Entry):
            if toggle_value:
                self.favorite_badge.setHidden(False)
                entry.add_tag(
                    self.panel.driver.lib,
                    tag_id,
                    field_id=FieldID.META_TAGS,
                    field_index=-1,
                )
            else:
                entry.remove_tag(self.panel.driver.lib, tag_id)

        # Is the badge a part of the selection?
        if (ItemType.ENTRY, self.item_id) in self.panel.driver.selected:
            # Yes, add chosen tag to all selected.
            for _, item_id in self.panel.driver.selected:
                entry = self.lib.get_entry(item_id)
                toggle_tag(entry)
        else:
            # No, add tag to the entry this badge is on.
            entry = self.lib.get_entry(self.item_id)
            toggle_tag(entry)

        if self.panel.isOpen:
            self.panel.update_widgets()
        self.panel.driver.update_badges()



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

"""TagStudio launcher."""

from src.core.ts_core import TagStudioCore
from src.cli.ts_cli import CliDriver  # type: ignore
from src.qt.ts_qt import QtDriver
import argparse
import traceback


def main():
    # appid = "cyanvoxel.tagstudio.9"
    # ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)

    # Parse arguments.
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--open",
        dest="open",
        type=str,
        help="Path to a TagStudio Library folder to open on start.",
    )
    parser.add_argument(
        "-o",
        dest="open",
        type=str,
        help="Path to a TagStudio Library folder to open on start.",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        type=str,
        help="Path to a TagStudio .ini or .plist config file to use.",
    )

    # parser.add_argument('--browse', dest='browse', action='store_true',
    #                     help='Jumps to entry browsing on startup.')
    # parser.add_argument('--external_preview', dest='external_preview', action='store_true',
    #                     help='Outputs current preview thumbnail to a live-updating file.')
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Reveals additional internal data useful for debugging.",
    )
    parser.add_argument(
        "--ui",
        dest="ui",
        type=str,
        help="User interface option for TagStudio. Options: qt, cli (Default: qt)",
    )
    parser.add_argument(
        "--ci",
        action=argparse.BooleanOptionalAction,
        help="Exit the application after checking it starts without any problem. Meant for CI check.",
    )
    args = parser.parse_args()

    core = TagStudioCore()  # The TagStudio Core instance. UI agnostic.
    driver = None  # The UI driver instance.
    ui_name: str = "unknown"  # Display name for the UI, used in logs.

    # Driver selection based on parameters.
    if args.ui and args.ui == "qt":
        driver = QtDriver(core, args)
        ui_name = "Qt"
    elif args.ui and args.ui == "cli":
        driver = CliDriver(core, args)
        ui_name = "CLI"
    else:
        driver = QtDriver(core, args)
        ui_name = "Qt"

    # Run the chosen frontend driver.
    try:
        driver.start()
    except Exception:
        traceback.print_exc()
        print(f"\nTagStudio Frontend ({ui_name}) Crashed! Press Enter to Continue...")
        input()


if __name__ == "__main__":
    main()



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from PySide6.QtCore import Signal, Qt
from PySide6.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QWidget,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QStyledItemDelegate,
    QLineEdit,
    QComboBox,
    QLabel,
)

from src.core.library import Library
from src.qt.widgets.panel import PanelWidget


class FileExtensionItemDelegate(QStyledItemDelegate):
    def setModelData(self, editor, model, index):
        if isinstance(editor, QLineEdit):
            if editor.text() and not editor.text().startswith("."):
                editor.setText(f".{editor.text()}")
        super().setModelData(editor, model, index)


class FileExtensionModal(PanelWidget):
    done = Signal()

    def __init__(self, library: "Library"):
        super().__init__()
        # Initialize Modal =====================================================
        self.lib = library
        self.setWindowTitle("File Extensions")
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setMinimumSize(240, 400)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 6, 6, 6)

        # Create Table Widget --------------------------------------------------
        self.table = QTableWidget(len(self.lib.ext_list), 1)
        self.table.horizontalHeader().setVisible(False)
        self.table.verticalHeader().setVisible(False)
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.setItemDelegate(FileExtensionItemDelegate())

        # Create "Add Button" Widget -------------------------------------------
        self.add_button = QPushButton()
        self.add_button.setText("&Add Extension")
        self.add_button.clicked.connect(self.add_item)
        self.add_button.setDefault(True)
        self.add_button.setMinimumWidth(100)

        # Create Mode Widgets --------------------------------------------------
        self.mode_widget = QWidget()
        self.mode_layout = QHBoxLayout(self.mode_widget)
        self.mode_layout.setContentsMargins(0, 0, 0, 0)
        self.mode_layout.setSpacing(12)
        self.mode_label = QLabel()
        self.mode_label.setText("List Mode:")
        self.mode_combobox = QComboBox()
        self.mode_combobox.setEditable(False)
        self.mode_combobox.addItem("Exclude")
        self.mode_combobox.addItem("Include")
        self.mode_combobox.setCurrentIndex(0 if self.lib.is_exclude_list else 1)
        self.mode_combobox.currentIndexChanged.connect(
            lambda i: self.update_list_mode(i)
        )
        self.mode_layout.addWidget(self.mode_label)
        self.mode_layout.addWidget(self.mode_combobox)
        self.mode_layout.setStretch(1, 1)

        # Add Widgets To Layout ------------------------------------------------
        self.root_layout.addWidget(self.mode_widget)
        self.root_layout.addWidget(self.table)
        self.root_layout.addWidget(
            self.add_button, alignment=Qt.AlignmentFlag.AlignCenter
        )

        # Finalize Modal -------------------------------------------------------
        self.refresh_list()

    def update_list_mode(self, mode: int):
        """
        Update the mode of the extension list: "Exclude" or "Include".

        Args:
            mode (int): The list mode, given by the index of the mode inside
                the mode combobox. 0 for "Exclude", 1 for "Include".
        """
        if mode == 0:
            self.lib.is_exclude_list = True
        elif mode == 1:
            self.lib.is_exclude_list = False

    def refresh_list(self):
        for i, ext in enumerate(self.lib.ext_list):
            self.table.setItem(i, 0, QTableWidgetItem(ext))

    def add_item(self):
        self.table.insertRow(self.table.rowCount())

    def save(self):
        self.lib.ext_list.clear()
        for i in range(self.table.rowCount()):
            ext = self.table.item(i, 0)
            if ext and ext.text():
                self.lib.ext_list.append(ext.text().lower())



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

from PySide6.QtCore import Signal, Qt, QSize
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLineEdit,
    QScrollArea,
    QFrame,
)

from src.core.library import Library
from src.qt.widgets.panel import PanelWidget, PanelModal
from src.qt.widgets.tag import TagWidget
from src.qt.modals.build_tag import BuildTagPanel


class TagDatabasePanel(PanelWidget):
    tag_chosen = Signal(int)

    def __init__(self, library):
        super().__init__()
        self.lib: Library = library
        # self.callback = callback
        self.first_tag_id = -1
        self.tag_limit = 30
        # self.selected_tag: int = 0

        self.setMinimumSize(300, 400)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 0, 6, 0)

        self.search_field = QLineEdit()
        self.search_field.setObjectName("searchField")
        self.search_field.setMinimumSize(QSize(0, 32))
        self.search_field.setPlaceholderText("Search Tags")
        self.search_field.textEdited.connect(
            lambda x=self.search_field.text(): self.update_tags(x)
        )
        self.search_field.returnPressed.connect(
            lambda checked=False: self.on_return(self.search_field.text())
        )

        # self.content_container = QWidget()
        # self.content_layout = QHBoxLayout(self.content_container)

        self.scroll_contents = QWidget()
        self.scroll_layout = QVBoxLayout(self.scroll_contents)
        self.scroll_layout.setContentsMargins(6, 0, 6, 0)
        self.scroll_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        self.scroll_area = QScrollArea()
        # self.scroll_area.setStyleSheet('background: #000000;')
        self.scroll_area.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOn
        )
        # self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setFrameShadow(QFrame.Shadow.Plain)
        self.scroll_area.setFrameShape(QFrame.Shape.NoFrame)
        # sa.setMaximumWidth(self.preview_size[0])
        self.scroll_area.setWidget(self.scroll_contents)

        # self.add_button = QPushButton()
        # self.root_layout.addWidget(self.add_button)
        # self.add_button.setText('Add Tag')
        # # self.done_button.clicked.connect(lambda checked=False, x=1101: (callback(x), self.hide()))
        # self.add_button.clicked.connect(lambda checked=False, x=1101: callback(x))
        # # self.setLayout(self.root_layout)

        self.root_layout.addWidget(self.search_field)
        self.root_layout.addWidget(self.scroll_area)
        self.update_tags("")

    # def reset(self):
    # 	self.search_field.setText('')
    # 	self.update_tags('')
    # 	self.search_field.setFocus()

    def on_return(self, text: str):
        if text and self.first_tag_id >= 0:
            # callback(self.first_tag_id)
            self.search_field.setText("")
            self.update_tags("")
        else:
            self.search_field.setFocus()
            self.parentWidget().hide()

    def update_tags(self, query: str):
        # TODO: Look at recycling rather than deleting and reinitializing
        while self.scroll_layout.itemAt(0):
            self.scroll_layout.takeAt(0).widget().deleteLater()

        # If there is a query, get a list of tag_ids that match, otherwise return all
        if query:
            tags = self.lib.search_tags(query, include_cluster=True)[
                : self.tag_limit - 1
            ]
        else:
            # Get tag ids to keep this behaviorally identical
            tags = [t.id for t in self.lib.tags]

        first_id_set = False
        for tag_id in tags:
            if not first_id_set:
                self.first_tag_id = tag_id
                first_id_set = True
            container = QWidget()
            row = QHBoxLayout(container)
            row.setContentsMargins(0, 0, 0, 0)
            row.setSpacing(3)
            tw = TagWidget(self.lib, self.lib.get_tag(tag_id), True, False)
            tw.on_edit.connect(
                lambda checked=False, t=self.lib.get_tag(tag_id): (self.edit_tag(t.id))
            )
            row.addWidget(tw)
            self.scroll_layout.addWidget(container)

        self.search_field.setFocus()

    def edit_tag(self, tag_id: int):
        btp = BuildTagPanel(self.lib, tag_id)
        # btp.on_edit.connect(lambda x: self.edit_tag_callback(x))
        self.edit_modal = PanelModal(
            btp,
            self.lib.get_tag(tag_id).display_name(self.lib),
            "Edit Tag",
            done_callback=(self.update_tags(self.search_field.text())),
            has_save=True,
        )
        # self.edit_modal.widget.update_display_name.connect(lambda t: self.edit_modal.title_widget.setText(t))
        # TODO Check Warning: Expected type 'BuildTagPanel', got 'PanelWidget' instead
        self.edit_modal.saved.connect(lambda: self.edit_tag_callback(btp))
        self.edit_modal.show()

    def edit_tag_callback(self, btp: BuildTagPanel):
        self.lib.update_tag(btp.build_tag())
        self.update_tags(self.search_field.text())

    # def enterEvent(self, event: QEnterEvent) -> None:
    # 	self.search_field.setFocus()
    # 	return super().enterEvent(event)
    # 	self.focusOutEvent



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from PySide6 import QtCore
from PySide6.QtCore import QEvent
from PySide6.QtGui import QEnterEvent, QPainter, QColor, QPen, QPainterPath, QPaintEvent
from PySide6.QtWidgets import QWidget
from src.qt.helpers.qbutton_wrapper import QPushButtonWrapper


class ThumbButton(QPushButtonWrapper):
    def __init__(self, parent: QWidget, thumb_size: tuple[int, int]) -> None:
        super().__init__(parent)
        self.thumb_size: tuple[int, int] = thumb_size
        self.hovered = False
        self.selected = False

        # self.clicked.connect(lambda checked: self.set_selected(True))

    def paintEvent(self, event: QPaintEvent) -> None:
        super().paintEvent(event)
        if self.hovered or self.selected:
            painter = QPainter()
            painter.begin(self)
            painter.setRenderHint(QPainter.RenderHint.Antialiasing)
            # painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)
            path = QPainterPath()
            width = 3
            radius = 6
            path.addRoundedRect(
                QtCore.QRectF(
                    width / 2,
                    width / 2,
                    self.thumb_size[0] - width,
                    self.thumb_size[1] - width,
                ),
                radius,
                radius,
            )

            # color = QColor('#bb4ff0') if self.selected else QColor('#55bbf6')
            # pen = QPen(color, width)
            # painter.setPen(pen)
            # # brush.setColor(fill)
            # painter.drawPath(path)

            if self.selected:
                painter.setCompositionMode(
                    QPainter.CompositionMode.CompositionMode_HardLight
                )
                color = QColor("#bb4ff0")
                color.setAlphaF(0.5)
                pen = QPen(color, width)
                painter.setPen(pen)
                painter.fillPath(path, color)
                painter.drawPath(path)

                painter.setCompositionMode(
                    QPainter.CompositionMode.CompositionMode_Source
                )
                color = QColor("#bb4ff0") if not self.hovered else QColor("#55bbf6")
                pen = QPen(color, width)
                painter.setPen(pen)
                painter.drawPath(path)
            elif self.hovered:
                painter.setCompositionMode(
                    QPainter.CompositionMode.CompositionMode_Source
                )
                color = QColor("#55bbf6")
                pen = QPen(color, width)
                painter.setPen(pen)
                painter.drawPath(path)
            painter.end()

    def enterEvent(self, event: QEnterEvent) -> None:
        self.hovered = True
        self.repaint()
        return super().enterEvent(event)

    def leaveEvent(self, event: QEvent) -> None:
        self.hovered = False
        self.repaint()
        return super().leaveEvent(event)

    def set_selected(self, value: bool) -> None:
        self.selected = value
        self.repaint()
